#!/usr/bin/env python
import tempfile, os
from . import custom_arguments
from .argparse_compat import argparse
from .import_util import lib_of_filename
from .diagnose_error import get_coq_output
from .import_util import lib_of_filename, norm_libname
from .import_util import has_dir_binding, deduplicate_trailing_dir_bindings
from .coq_version import (
    get_coqc_version,
    get_coqc_help,
    get_coq_accepts_top,
    DEFAULT_COQTOP,
)
from .custom_arguments import (
    add_libname_arguments,
    update_env_with_libnames,
    add_logging_arguments,
    process_logging_arguments,
    get_parser_name_mapping,
    LOG_ALWAYS,
)
from .binding_util import process_maybe_list
from .file_util import clean_v_file
from . import util
from .util import PY3

if PY3:
    from .util import raw_input

__all__ = ["main"]

parser = custom_arguments.ArgumentParser(
    description="List all identifiers which are not closed under the global context"
)
parser.add_argument(
    "--coqbin",
    metavar="COQBIN",
    dest="coqbin",
    type=str,
    default="",
    help="The path to a folder containing the coqc and coqtop programs.",
)
parser.add_argument(
    "--coqc",
    metavar="COQC",
    dest="coqc",
    type=str,
    default=None,
    action="append",
    help="The path to the coqc program.",
)
parser.add_argument(
    "--coqc-is-coqtop",
    dest="coqc_is_coqtop",
    default=False,
    action="store_const",
    const=True,
    help="Strip the .v and pass -load-vernac-source to the coqc programs; this allows you to pass `--coqc coqtop'",
)
parser.add_argument(
    "--coqtop",
    metavar="COQTOP",
    dest="coqtop",
    type=str,
    default=None,
    action="append",
    help=("The path to the coqtop program (default: %s)." % DEFAULT_COQTOP),
)
add_libname_arguments(parser)
add_logging_arguments(parser)


def qualify_identifiers_helper(identifiers, keep_unfound=False, **kwargs):
    for ident in identifiers:
        found_file = False
        if "." in ident:
            parts = ident.split(".")
            for i in range(len(parts) - 1, 0, -1):
                if not found_file:
                    old_libname = ".".join(parts[:i])
                    new_libname = norm_libname(old_libname, **kwargs)
                    if old_libname != new_libname:
                        yield (ident, new_libname + "." + ".".join(parts[i:]))
                        found_file = True
        if keep_unfound and not found_file:
            yield (ident, ident)


def qualify_identifiers(identifiers, keep_unfound=False, **kwargs):
    for ident, new_ident in qualify_identifiers_helper(
        identifiers, keep_unfound=keep_unfound, **kwargs
    ):
        yield new_ident


def filter_local_identifiers(identifiers, keep_unfound=False, **kwargs):
    for ident, new_ident in qualify_identifiers_helper(
        identifiers, keep_unfound=keep_unfound, **kwargs
    ):
        yield ident


def get_constant_name_from_locate(val):
    val = val.strip()
    for ctype in ("Constant", "Constructor", "Inductive", ""):
        if val and val[: len(ctype)] == ctype:
            val = val[len(ctype) :].strip()
            if "\n" in val:
                val = val[: val.index("\n")].strip()
            return (val, ctype)


def fix_identifiers(idents, libname):
    for ident in idents:
        if ident[: len("Warning:")] == "Warning:":
            raw_input(ident)
            continue
        elif "." not in ident:
            yield ident
        else:
            idx = ident.index(".")
            overlap, rest = ident[:idx], ident[idx:]
            if libname[-(idx + 1) :] == "." + overlap:
                yield libname + rest
            else:
                # raw_input((ident, libname))
                # yield ident
                continue


def main():
    try:
        args = process_logging_arguments(parser.parse_args())
    except argparse.ArgumentError as exc:
        if exc.message == "expected one argument":
            exc.reraise(
                '\nNote that argparse does not accept arguments with leading dashes.\nTry --foo=bar or --foo " -bar", if this was your intent.\nSee Python issue 9334.'
            )
        else:
            exc.reraise()

    def prepend_coqbin(prog):
        if isinstance(prog, str):
            prog = (prog,)
        if args.coqbin != "":
            return (os.path.join(args.coqbin, prog[0]), *prog[1:])
        else:
            return tuple(prog)

    env = {
        "log": args.log,
        "coqc": prepend_coqbin(args.coqc or "coqc"),
        "coqtop": prepend_coqbin(args.coqtop or DEFAULT_COQTOP),
        "coqc_args": tuple(
            i.strip() for i in process_maybe_list(args.coq_args, log=args.log)
        ),
        "coqc_is_coqtop": args.coqc_is_coqtop,
        "temp_file_name": "",
        "cli_mapping": get_parser_name_mapping(parser),
    }

    env["remove_temp_file"] = False
    if env["temp_file_name"] == "":
        temp_file = tempfile.NamedTemporaryFile(suffix=".v", dir=".", delete=False)
        env["temp_file_name"] = temp_file.name
        temp_file.close()
        env["remove_temp_file"] = True

    if env["coqc_is_coqtop"]:
        if env["coqc"] == ("coqc",):
            env["coqc"] = env["coqtop"]
        env["make_coqc"] = (util.resource_path("coqtop-as-coqc.sh"), *env["coqc"])

    coqc_help = get_coqc_help(env["coqc"], **env)
    coqc_version = get_coqc_version(env["coqc"], **env)

    if has_dir_binding(env["coqc_args"], coqc_help=coqc_help):
        update_env_with_libnames(env, args, default=tuple([]))
    else:
        update_env_with_libnames(env, args)

    for dirname, libname in env["libnames"]:
        env["coqc_args"] = tuple(list(env["coqc_args"]) + ["-R", dirname, libname])
    for dirname, libname in env["non_recursive_libnames"]:
        env["coqc_args"] = tuple(list(env["coqc_args"]) + ["-Q", dirname, libname])
    for dirname in env["ocaml_dirnames"]:
        env["coqc_args"] = tuple(list(env["coqc_args"]) + ["-I", dirname])
    env["coqc_args"] = deduplicate_trailing_dir_bindings(
        env["coqc_args"],
        coqc_help=coqc_help,
        coq_accepts_top=get_coq_accepts_top(env["coqc"]),
    )

    try:
        env["log"]("Listing identifiers...")
        unknown = []
        closed_idents = []
        open_idents = []
        errors = []
        ignore_header = ("Welcome to Coq", "[Loading ML file")
        error_header = ("Toplevel input, characters",)
        for filename in sorted(env["_CoqProject_v_files"]):
            env["log"]("Qualifying identifiers in %s..." % filename, level=2)
            env["log"]("Printing assumptions in %s..." % filename, max_level=2)
            libname = lib_of_filename(filename, **env)
            require_statement = "Require %s.\n" % libname
            search_code = r"""%s
Set Search Output Name Only.
SearchPattern _ inside %s.""" % (
                require_statement,
                libname,
            )
            output, cmds, retcode, runtime, _peak_rss_kb = get_coq_output(
                env["coqc"],
                env["coqc_args"],
                search_code,
                0,
                is_coqtop=env["coqc_is_coqtop"],
                verbose_base=3,
                **env,
            )
            identifiers = sorted(
                fix_identifiers(
                    set(i.strip() for i in output.split("\n") if i.strip()), libname
                )
            )
            print_assumptions_code = require_statement + "\n".join(
                "Locate %s.\nPrint Assumptions %s." % (i, i) for i in identifiers
            )
            env["log"]("Printing assumptions...", level=2)
            output, cmds, retcode, runtime, _peak_rss_kb = get_coq_output(
                env["coqtop"],
                env["coqc_args"],
                print_assumptions_code,
                0,
                is_coqtop=True,
                pass_on_stdin=True,
                verbose_base=3,
                **env,
            )
            i = 0
            statements = output.split("\nCoq <")
            while i < len(statements):
                if i + 1 < len(
                    statements
                ) and " Closed under the global context " in statements[i + 1].replace(
                    "\n", " "
                ):
                    last, ctype = get_constant_name_from_locate(statements[i])
                    closed_idents.append((last, ctype))
                    env["log"]("Closed: %s (%s)" % (last, ctype), level=2)
                    i += 2
                elif i + 1 < len(statements) and "Axioms:" in statements[i + 1].replace(
                    "\n", " "
                ):
                    last, ctype = get_constant_name_from_locate(statements[i])
                    open_idents.append((last, ctype, statements[i + 1]))
                    env["log"]("OPEN: %s (%s)" % (last, ctype))
                    env["log"](statements[i + 1], level=3)
                    i += 2
                elif not statements[i].strip():
                    i += 1
                else:
                    found_ignore, found_error = False, False
                    for header in ignore_header:
                        if statements[i].strip()[: len(header)] == header:
                            found_ignore = True
                            env["log"]("Ignoring: %s" % statements[i], level=3)
                    for header in ignore_header:
                        if statements[i].strip()[: len(header)] == header:
                            found_error = True
                    if not found_ignore:
                        if found_error:
                            errors.append(statements[i])
                            env["log"]("ERROR: %s" % statements[i])
                        else:
                            env["log"]("UNKNOWN: %s" % statements[i])
                            unknown.append(statements[i])
                    i += 1
        env["log"](
            "Identifiers which are not closed under the global context:\n%s"
            % "\n".join(ident[0] for ident in open_idents),
            level=LOG_ALWAYS,
        )
    finally:
        if env["remove_temp_file"]:
            clean_v_file(env["temp_file_name"])


if __name__ == "__main__":
    main()



import os
import re
import shlex
import shutil
import sys
import subprocess
from difflib import SequenceMatcher
from typing import Dict, Hashable, Iterable, List, Set, TypeVar, Tuple

try:
    from typing import Protocol
except ImportError:
    Protocol = None

from .argparse_compat import argparse

__all__ = [
    "prompt",
    "yes_no_prompt",
    "b",
    "s",
    "cmp_compat",
    "PY3",
    "BooleanOptionalAction",
    "argparse",
    "raw_input",
    "re_escape",
    "slice_string_at_bytes",
    "len_in_bytes",
    "shlex_quote",
    "shlex_join",
    "resource_path",
    "group_by",
    "transitive_closure",
    "get_peak_rss_bytes",
    "parse_memory_bytes",
    "wrap_with_prlimit",
    "wrap_with_ulimit",
    "wrap_with_cgexec",
    "wrap_with_cgexec_and_create",
    "wrap_with_systemd_run",
    "cleanup_cgroup",
    "apply_memory_limit",
    "MEMORY_LIMIT_METHODS",
]

BooleanOptionalAction = argparse.BooleanOptionalAction

SCRIPT_DIRECTORY = os.path.dirname(os.path.realpath(__file__))

if sys.version_info < (3,):
    PY3 = False

    def b(x):
        return x

    def s(x):
        return x

else:
    PY3 = True

    def b(x) -> bytes:
        if x is not None:
            return x.encode()
        return x

    def s(x) -> str:
        # Sometimes we get str rather than bytes??? cf https://gitlab.com/coq/coq/-/jobs/544269051
        if hasattr(x, "decode"):
            return x.decode("utf-8", "ignore")
        return x

    def cmp(x, y):
        if x < y:
            return -1
        if y < x:
            return 1
        return 0

    raw_input = input

if sys.version_info < (3, 7):
    # see https://bugs.python.org/issue29995
    def re_escape(*args, **kwargs):
        ret = re.escape(*args, **kwargs)
        for ch in ':"':
            ret = ret.replace("\\" + ch, ch)
        return ret

else:
    re_escape = re.escape

cmp_compat = cmp

K = TypeVar("K")
V = TypeVar("V")
T = TypeVar("T")
TH = TypeVar("TH", bound=Hashable)


def prompt(
    options,
    case_sensitive=False,
    strip=True,
    sep="/",
    prefix="Please enter ",
    postfix=": ",
    yes_value=True,
    yes=False,
):
    msg = prefix + sep.join(i["display"] for i in options) + postfix
    if yes:
        print(msg)
        return yes_value
    else:
        response = raw_input(msg)
        while True:
            if not case_sensitive:
                response = response.lower()
            if strip:
                response = response.strip()
            for expected in options:
                if response in expected["values"]:
                    return expected["value"]
            print(
                'Response "%s" is not one of %s'
                % (
                    response,
                    ", ".join(repr(val) for i in options for val in i["values"]),
                )
            )
            response = raw_input(msg)


def yes_no_prompt(**kwargs):
    return prompt(
        (
            {"value": True, "display": "(y)es", "values": ("y", "yes")},
            {"value": False, "display": "(n)o", "values": ("n", "no")},
        ),
        yes_value=True,
        **kwargs,
    )


# Unfortunately, coqtop -emacs -time reports character locations in
# bytes, as does the glob file, so we need to handle unicode.  Here we
# write a generic slicer based on unicode locations that works across
# versions of python
def slice_string_at_bytes(string, start=None, end=None):
    string = b(string)
    if start is None:
        start = 0
    if end is None:
        end = len(string)
    return s(string[start:end])


def len_in_bytes(string):
    return len(b(string))


def normalize_newlines(string: str) -> str:
    return string.replace("\r\n", "\n").replace("\r", "\n")


# Terminal colors (maybe something cleverer needs to be done for other
# platforms).
class colors:
    ESC = "\033"
    # Escape code doesn't render on github so we use the standard escaped escape.
    # ESC = ''

    HEADER = ESC + "[95m"
    OKBLUE = ESC + "[94m"
    OKCYAN = ESC + "[96m"
    OKGREEN = ESC + "[92m"
    WARNING = ESC + "[93m"
    FAIL = ESC + "[91m"
    ENDC = ESC + "[0m"
    BOLD = ESC + "[1m"
    UNDERLINE = ESC + "[4m"


# Colors a string a given color
# Example usage: color("Hello World!", colors.OKBLUE)
def color(str, col, on=True):
    if not on:
        return str
    else:
        return col + str + colors.ENDC


if sys.version_info < (3, 3):
    import pipes

    shlex_quote = pipes.quote
else:
    import shlex

    shlex_quote = shlex.quote

if sys.version_info < (3, 8):
    shlex_join = lambda split_command: " ".join(
        shlex_quote(arg) for arg in split_command
    )
else:
    import shlex

    shlex_join = shlex.join


def resource_path(path):
    if os.path.exists(os.path.join(SCRIPT_DIRECTORY, path)):
        return os.path.join(SCRIPT_DIRECTORY, path)
    if sys.version_info < (3, 7):
        import pkg_resources

        return pkg_resources.resource_filename("coq_tools", path)
    else:
        import importlib.resources

        return importlib.resources.path("coq_tools", path)


def group_by_iter(ls, f):
    """
    Groups elements in a list based on a given condition.

    Args:
        ls (iterable): The list of elements to be grouped.
        f (function): The condition function used for grouping.

    Returns:
        iterable: A list of lists, where each inner list represents a group of elements.
    """
    it = iter(ls)
    prev = next(it)
    cur = [prev]
    for token in it:
        if f(prev, token):
            cur.append(token)
        else:
            yield cur
            cur = [token]
        prev = token
    yield cur


def group_by(ls, f):
    """
    Groups elements in a list based on a given condition.

    Args:
        ls (list): The list of elements to be grouped.
        f (function): The condition function used for grouping.

    Returns:
        list: A list of lists, where each inner list represents a group of elements.

    Example:
        >>> group_by([1, 2, 3, 4, 5, 7, 8, 10], lambda x, y: x + 1 == y)
        [[1, 2, 3, 4, 5], [7, 8], [10]]
    """
    return list(group_by_iter(ls, f))


def list_diff(
    old_strs: List[str], new_strs: List[str], *, autojunk: bool = False
) -> str:
    """
    >>> old = ["line1", "line2", "line3", "line4"]
    >>> new = ["line1", "line3", "line4", "line5"]
    >>> print(list_diff(old, new))
      line1
    - line2
      line3
      line4
    + line5
    """
    # Create a SequenceMatcher object
    sm = SequenceMatcher(None, old_strs, new_strs, autojunk=autojunk)
    opcodes = sm.get_opcodes()
    diff_lines = []

    # Threshold for eliding long equal blocks (if more than this many lines, collapse)
    ELIDE_THRESHOLD = 3

    for tag, i1, i2, j1, j2 in opcodes:
        if tag == "equal":
            block_length = i2 - i1
            if block_length > ELIDE_THRESHOLD:
                # Show the first line, an ellipsis, then the last line of the block.
                diff_lines.append("  " + old_strs[i1])
                diff_lines.append("  ...")
                diff_lines.append("  " + old_strs[i2 - 1])
            else:
                for i in range(i1, i2):
                    diff_lines.append("  " + old_strs[i])
        elif tag == "delete":
            # Lines in old_strs that were removed.
            for i in range(i1, i2):
                diff_lines.append("- " + old_strs[i])
        elif tag == "insert":
            # Lines added in new_strs.
            for j in range(j1, j2):
                diff_lines.append("+ " + new_strs[j])
        elif tag == "replace":
            # Show removed lines
            for i in range(i1, i2):
                diff_lines.append("- " + old_strs[i])
            # Then show inserted lines
            for j in range(j1, j2):
                diff_lines.append("+ " + new_strs[j])

    return "\n".join(diff_lines)


# if Protocol is not None:

#     class _TransitiveClosureDict(Generic[TH], Protocol):
#         def __getitem__(self, key: TH) -> Iterable[TH]: ...

#         def keys(self) -> Iterable[TH]: ...

#         def values(self) -> Iterable[Iterable[TH]]: ...

#     TransitiveClosureDict = _TransitiveClosureDict
# else:

#     class _TransitiveClosureDictOld(Generic[TH], Dict[TH, Iterable[TH]]):
#         pass

#     TransitiveClosureDict = _TransitiveClosureDictOld


def transitive_closure(
    graph: Dict[TH, Iterable[TH]],
) -> Dict[TH, Set[TH]]:
    """
    Computes the transitive closure of a directed graph represented as an adjacency list.

    The input is a dictionary where keys are nodes and values are iterables of neighboring nodes.
    The output is a dictionary where each key is a node and the value is the set of all reachable nodes
    from that key, including itself.

    This implementation uses iterative DFS for each starting node to handle large graphs and avoid recursion depth issues.
    """
    # Collect all nodes (keys and those appearing in values)
    all_nodes = set(graph.keys())
    for neighbors in graph.values():
        all_nodes.update(neighbors)

    def adj(node):
        # allow fills from __missing__
        try:
            return graph[node]
        except KeyError:
            return ()

    closure = {}
    for start in all_nodes:
        visited = set()
        stack = [start]
        while stack:
            node = stack.pop()
            if node not in visited:
                visited.add(node)
                for neighbor in adj(node):
                    if neighbor not in visited:
                        stack.append(neighbor)
        closure[start] = visited

    return closure


def get_peak_rss_bytes(rusage):
    """
    Extract peak RSS (Resident Set Size) from rusage in bytes.

    Args:
        rusage: A resource usage object from os.wait4() or similar.

    Returns:
        Peak RSS in bytes. On Linux, ru_maxrss is in kilobytes, so we multiply by 1024.
        On macOS/BSD, ru_maxrss is already in bytes.
    """
    if rusage is None:
        return None
    if sys.platform.startswith("linux"):
        return rusage.ru_maxrss * 1024
    else:
        return rusage.ru_maxrss


def parse_memory_bytes(mem_str: str) -> int:
    """
    Parse a memory string (case-insensitive) to bytes.

    Supports formats like: "5M", "5G", "5GB", "5GiB", "512", "unlimited", "0"

    Args:
        mem_str: Memory string to parse (case-insensitive)

    Returns:
        Number of bytes, or -1 for "unlimited" or "0"

    Examples:
        >>> parse_memory_bytes("5M")
        5242880
        >>> parse_memory_bytes("5G")
        5368709120
        >>> parse_memory_bytes("5GB")
        5368709120
        >>> parse_memory_bytes("5GiB")
        5368709120
        >>> parse_memory_bytes("512")
        512
        >>> parse_memory_bytes("unlimited")
        -1
        >>> parse_memory_bytes("0")
        -1
    """
    if isinstance(mem_str, (int, float)):
        val = float(mem_str)
        return int(val) if val > 0 else -1

    mem_str_normalized = str(mem_str).strip().lower()
    mem_str = mem_str_normalized

    if mem_str in ("unlimited", "0", ""):
        return -1

    # Match number with optional unit suffix
    # Pattern: number followed by optional unit
    # Units: k/m/g/t (with optional 'i' for binary, optional 'b' suffix)
    # Examples: "5", "5k", "5kb", "5kib", "5m", "5mb", "5mib", "5g", "5gb", "5gib"
    match = re.match(r"^(\d+(?:\.\d+)?)\s*([kmgt]i?b?)$", mem_str)
    if not match:
        # Try to parse as plain number
        try:
            val = int(mem_str)
            return val if val > 0 else -1
        except ValueError:
            raise ValueError(f"Invalid memory format: {mem_str}")

    number_str, unit = match.groups()
    number = float(number_str)

    # Parse unit - handle both "b" and "ib" suffixes
    unit = unit.lower()
    if unit == "" or unit == "b":
        multiplier = 1
    elif unit in ("k", "kb"):
        multiplier = 1000
    elif unit == "kib":
        multiplier = 1024
    elif unit in ("m", "mb"):
        multiplier = 1000**2
    elif unit == "mib":
        multiplier = 1024**2
    elif unit in ("g", "gb"):
        multiplier = 1000**3
    elif unit == "gib":
        multiplier = 1024**3
    elif unit in ("t", "tb"):
        multiplier = 1000**4
    elif unit == "tib":
        multiplier = 1024**4
    else:
        raise ValueError(f"Unknown memory unit: {unit}")

    return int(number * multiplier)


def wrap_with_prlimit(
    cmd: List[str], as_bytes: int = None, rss_bytes: int = None
) -> List[str]:
    """
    Wrap command with prlimit to set resource limits.

    Args:
        cmd: Command to run
        as_bytes: RLIMIT_AS (virtual memory) limit in bytes
        rss_bytes: RLIMIT_RSS limit in bytes (advisory on Linux)

    Returns:
        Wrapped command list
    """
    if (as_bytes is None or as_bytes <= 0) and (rss_bytes is None or rss_bytes <= 0):
        return cmd

    wrapper = ["prlimit"]
    if as_bytes is not None and as_bytes > 0:
        wrapper.append(f"--as={as_bytes}")
    if rss_bytes is not None and rss_bytes > 0:
        wrapper.append(f"--rss={rss_bytes}")
    wrapper.append("--")
    return wrapper + cmd


def wrap_with_ulimit(cmd: List[str], as_bytes: int = None) -> List[str]:
    """
    Wrap command with shell ulimit to set RLIMIT_AS.

    Args:
        cmd: Command to run
        as_bytes: RLIMIT_AS (virtual memory) limit in bytes

    Returns:
        Wrapped command list using sh -c
    """
    if as_bytes is None or as_bytes <= 0:
        return cmd

    # ulimit -v takes kilobytes, use ceiling division to avoid exceeding limit
    kb = (as_bytes + 1023) // 1024
    escaped_cmd = " ".join(shlex.quote(arg) for arg in cmd)
    return ["sh", "-c", f"ulimit -v {kb}; exec {escaped_cmd}"]


def wrap_with_cgexec(
    cmd: List[str], cgroup: str, controllers: str = "memory"
) -> List[str]:
    """
    Wrap command with cgexec to run in an existing cgroup.

    Args:
        cmd: Command to run
        cgroup: Name of existing cgroup (e.g., "limited_group")
        controllers: Cgroup controllers to use (default: "memory")

    Returns:
        Wrapped command list

    Note:
        The cgroup must already exist and have limits configured.
        Example setup:
            sudo cgcreate -g memory:/limited_group
            echo 100M | sudo tee /sys/fs/cgroup/memory/limited_group/memory.limit_in_bytes
    """
    if not cgroup:
        return cmd
    return ["cgexec", "-g", f"{controllers}:{cgroup}"] + cmd


def parse_memory_limit_with_multiplier(value, arg_name):
    """returns (absolute_bytes_or_minus_one_or_none, multiplier_or_none)"""
    if value is None:
        return None, None
    value_str = str(value).strip()
    if not value_str:
        return None, None
    lower_val = value_str.lower()
    if lower_val.endswith("x"):
        multiplier_str = lower_val[:-1].strip()
        if not multiplier_str:
            raise ValueError(
                f"{arg_name} multiplier must include a numeric value before 'x'"
            )
        try:
            multiplier = float(multiplier_str)
        except ValueError as exc:
            raise ValueError(
                f"{arg_name} multiplier '{value_str}' is not a valid number"
            ) from exc
        if multiplier <= 0:
            raise ValueError(f"{arg_name} multiplier '{value_str}' must be positive")
        return None, multiplier
    try:
        return parse_memory_bytes(value_str), None
    except ValueError as exc:
        raise ValueError(f"{arg_name} has invalid value '{value_str}': {exc}") from exc


def wrap_with_cgexec_and_create(
    cmd: List[str],
    mem_bytes: int,
    swap_bytes: int = 0,
    cgroup_name: str = None,
) -> Tuple[List[str], str]:
    """
    Create a cgroup v2 with memory limits and wrap command with cgexec.

    Args:
        cmd: Command to run
        mem_bytes: Memory limit in bytes
        swap_bytes: Swap limit in bytes (0 = no swap, -1 = unlimited)
        cgroup_name: Optional cgroup name (auto-generated if None)

    Returns:
        Tuple of (wrapped command list, cgroup path for cleanup)

    Raises:
        RuntimeError: If cgroup creation fails
    """
    if mem_bytes is None or mem_bytes <= 0:
        return cmd, None

    cg_root = "/sys/fs/cgroup"
    cg_name = cgroup_name or f"pyjob-{os.getpid()}"
    cg_path = os.path.join(cg_root, cg_name)

    try:
        os.makedirs(cg_path, exist_ok=True)
        with open(os.path.join(cg_path, "memory.max"), "w") as f:
            f.write(str(mem_bytes))
        with open(os.path.join(cg_path, "memory.swap.max"), "w") as f:
            f.write(str(swap_bytes) if swap_bytes >= 0 else "max")
        with open(os.path.join(cg_path, "memory.oom.group"), "w") as f:
            f.write("1")
    except (OSError, IOError) as e:
        raise RuntimeError(f"Failed to set up cgroup at {cg_path}: {e}")

    wrapped_cmd = ["cgexec", "-g", f"memory:{cg_name}"] + cmd
    return wrapped_cmd, cg_path


def wrap_with_systemd_run(
    cmd: List[str],
    mem_bytes: int = None,
    as_bytes: int = None,
    swap_bytes: int = None,
    scope_name: str = None,
) -> List[str]:
    """
    Wrap command with systemd-run to set memory limits via cgroup.

    Args:
        cmd: Command to run
        mem_bytes: MemoryMax limit in bytes (hard RSS limit)
        as_bytes: LimitAS (virtual memory) limit in bytes
        swap_bytes: MemorySwapMax limit in bytes
        scope_name: Optional name for the transient scope unit

    Returns:
        Wrapped command list

    Note:
        Requires systemd and user session (--user) or root privileges.
    """
    has_limits = any(v is not None and v > 0 for v in [mem_bytes, as_bytes, swap_bytes])
    if not has_limits:
        return cmd

    wrapper = ["systemd-run", "--user", "--scope", "--quiet"]

    if scope_name:
        wrapper.append(f"--unit={scope_name}")

    if mem_bytes is not None and mem_bytes > 0:
        wrapper.append(f"--property=MemoryMax={mem_bytes}")

    if as_bytes is not None and as_bytes > 0:
        wrapper.append(f"--property=LimitAS={as_bytes}")

    if swap_bytes is not None and swap_bytes > 0:
        wrapper.append(f"--property=MemorySwapMax={swap_bytes}")

    wrapper.append("--")
    return wrapper + cmd


def cleanup_cgroup(cg_path: str) -> None:
    """
    Clean up a cgroup directory.

    Args:
        cg_path: Path to cgroup directory to remove
    """
    if cg_path is None:
        return
    try:
        shutil.rmtree(cg_path, ignore_errors=True)
    except Exception:
        pass


# Memory limit method enum for CLI
MEMORY_LIMIT_METHODS = ["prlimit", "ulimit", "cgexec", "systemd-run", "none"]


def apply_memory_limit(
    cmd: List[str],
    method: str,
    as_bytes: int = None,
    rss_bytes: int = None,
    swap_bytes: int = None,
    cgroup: str = None,
) -> Tuple[List[str], str]:
    """
    Apply memory limits to a command using the specified method.

    Args:
        cmd: Command to run
        method: One of "prlimit", "ulimit", "cgexec", "systemd-run", "none"
        as_bytes: Virtual memory (RLIMIT_AS) limit in bytes
        rss_bytes: RSS memory limit in bytes
        swap_bytes: Swap limit in bytes
        cgroup: Existing cgroup name (for cgexec method)

    Returns:
        Tuple of (wrapped command, cgroup_path or None)
        cgroup_path is only set when method creates a new cgroup that needs cleanup

    Raises:
        ValueError: If method is unknown or invalid configuration
    """
    method = method.lower() if method else "none"

    if method == "none":
        return cmd, None

    elif method == "prlimit":
        return wrap_with_prlimit(cmd, as_bytes=as_bytes, rss_bytes=rss_bytes), None

    elif method == "ulimit":
        if rss_bytes is not None and rss_bytes > 0:
            raise ValueError(
                "ulimit method only supports --max-mem-as, not --max-mem-rss"
            )
        return wrap_with_ulimit(cmd, as_bytes=as_bytes), None

    elif method == "cgexec":
        if cgroup:
            # Use existing cgroup
            return wrap_with_cgexec(cmd, cgroup), None
        elif rss_bytes is not None and rss_bytes > 0:
            # Create new cgroup with limits
            return wrap_with_cgexec_and_create(cmd, rss_bytes, swap_bytes or 0)
        else:
            raise ValueError("cgexec method requires --cgroup or --max-mem-rss")

    elif method == "systemd-run":
        return wrap_with_systemd_run(
            cmd, mem_bytes=rss_bytes, as_bytes=as_bytes, swap_bytes=swap_bytes
        ), None

    else:
        raise ValueError(
            f"Unknown memory limit method: {method}. "
            f"Valid options: {', '.join(MEMORY_LIMIT_METHODS)}"
        )


if __name__ == "__main__":
    # if we're working in Python 3.3, we can test this file
    try:
        import doctest

        success = True
    except ImportError:
        print(
            "This is not the main file to use.\nOnly run it if you have doctest (Python 3.3+) and are testing things."
        )
        success = False
    if success:
        doctest.testmod()



from __future__ import print_function, with_statement

import glob
import os
import os.path
import re
import shutil
import subprocess
import sys
import tempfile
import time
from collections import OrderedDict
from functools import cmp_to_key, partial
from pathlib import Path
from typing import Callable, Dict, Optional, Tuple

from . import util
from .coq_version import (
    coq_makefile_supports_arg,
    get_coq_accepts_o,
    get_coqc_help,
    group_coq_args,
    group_coq_args_split_recognized,
)
from .custom_arguments import DEFAULT_LOG, LOG_ALWAYS
from .memoize import memoize
from .split_file import get_coq_statement_byte_ranges, split_coq_file_contents
from .strip_comments import strip_comments, strip_trailing_comments
from .util import cmp_compat as cmp
from .util import shlex_join, shlex_quote, transitive_closure

__all__ = [
    "filename_of_lib",
    "lib_of_filename",
    "has_dir_binding",
    "deduplicate_trailing_dir_bindings",
    "deduplicate_trailing_dir_bindings_get_topname",
    "get_file_as_bytes",
    "get_file",
    "make_globs",
    "get_imports",
    "norm_libname",
    "recursively_get_imports",
    "IMPORT_ABSOLUTIZE_TUPLE",
    "ALL_ABSOLUTIZE_TUPLE",
    "absolutize_has_all_constants",
    "run_recursively_get_imports",
    "run_maybe_recursively_get_imports",
    "clear_libimport_cache",
    "get_byte_references_for",
    "sort_files_by_dependency",
    "get_recursive_requires",
    "get_recursive_require_names",
    "insert_references",
    "classify_require_kind",
    "REQUIRE",
    "REQUIRE_IMPORT",
    "REQUIRE_EXPORT",
    "EXPORT",
    "IMPORT",
    "get_references_from_globs",
    "split_requires_of_statements",
]

file_mtimes = {}
file_contents = {}
raw_file_contents = {}
lib_imports_fast = {}
lib_imports_slow = {}

DEFAULT_LIBNAMES = ((".", "Top"),)

IMPORT_ABSOLUTIZE_TUPLE = ("lib", "mod")
ALL_ABSOLUTIZE_TUPLE = (
    "lib",
    "proj",
    "rec",
    "ind",
    "constr",
    "def",
    "syndef",
    "class",
    "thm",
    "lem",
    "prf",
    "ax",
    "inst",
    "prfax",
    "coind",
    "scheme",
    "vardef",
    "mod",
)  # , 'modtype')

IMPORT_LINE_REG = re.compile(
    r"^\s*(?:Require\s+Import|Require\s+Export|Require|Load\s+Verbose|Load)\s+(.*?)\.(?:\s|$)",
    re.MULTILINE | re.DOTALL,
)


def warning(*objs):
    print("WARNING: ", *objs, file=sys.stderr)


def error(*objs):
    print("ERROR: ", *objs, file=sys.stderr)


def fill_kwargs(kwargs, for_makefile=False):
    rtn = {
        "libnames": DEFAULT_LIBNAMES,
        "non_recursive_libnames": tuple(),
        "ocaml_dirnames": tuple(),
        "log": DEFAULT_LOG,
        "coqc": ("coqc",),
        "coq_makefile": ("coq_makefile",),
        "coqdep": ("coqdep",),
        "use_coq_makefile_for_deps": True,
        "walk_tree": True,
        "coqc_args": tuple(),
        "cli_mapping": {
            "use_coq_makefile_for_deps": {False: ["--no-deps"]},
        },
        "coqpath_paths": tuple(),
    }
    rtn.update(kwargs)
    if for_makefile:
        if (
            "make_coqc" in rtn.keys()
        ):  # handle the case where coqc for the makefile is different
            rtn["coqc"] = rtn["make_coqc"]
        if (
            "passing_make_coqc" in rtn.keys()
        ):  # handle the case where coqc for the makefile is different
            rtn["passing_coqc"] = rtn["passing_make_coqc"]
    return rtn


def safe_kwargs(kwargs):
    for k, v in list(kwargs.items()):
        if isinstance(v, list):
            kwargs[k] = tuple(v)
    return dict((k, v) for k, v in kwargs.items() if not isinstance(v, dict))


def fix_path(filename):
    return filename.replace("\\", "/")


def absolutize_has_all_constants(absolutize_tuple):
    """Returns True if absolutizing the types of things mentioned by the tuple is enough to ensure that we only use absolute names"""
    return set(ALL_ABSOLUTIZE_TUPLE).issubset(set(absolutize_tuple))


def libname_with_dot(logical_name):
    if logical_name in ("", '""', "''"):
        return ""
    else:
        return logical_name + "."


def clear_libimport_cache(libname):
    if libname in lib_imports_fast.keys():
        del lib_imports_fast[libname]
    if libname in lib_imports_slow.keys():
        del lib_imports_slow[libname]


@memoize
def os_walk(top, topdown=True, onerror=None, followlinks=False):
    return tuple(
        os.walk(top, topdown=topdown, onerror=onerror, followlinks=followlinks)
    )


@memoize
def os_path_isfile(filename):
    return os.path.isfile(filename)


def filenames_of_lib_helper(lib, libnames, non_recursive_libnames, ext):
    for physical_name, logical_name in list(libnames) + list(non_recursive_libnames):
        if lib.startswith(libname_with_dot(logical_name)):
            cur_lib = lib[len(libname_with_dot(logical_name)) :]
            cur_lib = os.path.join(physical_name, cur_lib.replace(".", os.sep))
            yield fix_path(os.path.relpath(os.path.normpath(cur_lib + ext), "."))


def local_filenames_of_lib_helper(lib, libnames, non_recursive_libnames, ext):
    # is this the right thing to do?
    lib = lib.replace(".", os.sep)
    for dirpath, dirname, filenames in os_walk(".", followlinks=True):
        filename = os.path.relpath(
            os.path.normpath(os.path.join(dirpath, lib + ext)), "."
        )
        if os_path_isfile(filename):
            yield fix_path(filename)


@memoize
def filename_of_lib_helper(lib, libnames, non_recursive_libnames, ext):
    filenames = list(
        filenames_of_lib_helper(lib, libnames, non_recursive_libnames, ext)
    )
    # kludge for https://github.com/coq/coq/pull/19310
    if lib.startswith("Stdlib.") and not filenames:
        filenames = list(
            filenames_of_lib_helper(
                "Coq." + lib[len("Stdlib.") :], libnames, non_recursive_libnames, ext
            )
        )
        if filenames:
            DEFAULT_LOG(
                "WARNING: Using Coq in place of Stdlib when resolving %s to %s."
                % (lib, ", ".join(sorted(set(filenames)))),
                level=LOG_ALWAYS,
            )
    local_filenames = list(
        local_filenames_of_lib_helper(lib, libnames, non_recursive_libnames, ext)
    )
    existing_filenames = [
        f
        for f in filenames
        if os_path_isfile(f) or os_path_isfile(os.path.splitext(f)[0] + ".v")
    ]
    if len(existing_filenames) > 0:
        retval = existing_filenames[0]
        if len(set(existing_filenames)) == 1:
            return retval
        else:
            DEFAULT_LOG(
                "WARNING: Multiple physical paths match logical path %s: %s.  Selecting %s."
                % (lib, ", ".join(sorted(set(existing_filenames))), retval),
                level=LOG_ALWAYS,
            )
            return retval
    if len(filenames) != 0:
        DEFAULT_LOG(
            "WARNING: One or more physical paths match logical path %s, but none of them exist: %s"
            % (lib, ", ".join(filenames)),
            level=LOG_ALWAYS,
        )
    if len(local_filenames) > 0:
        retval = local_filenames[0]
        if len(local_filenames) == 1:
            return retval
        else:
            DEFAULT_LOG(
                "WARNING: Multiple local physical paths match logical path %s: %s.  Selecting %s."
                % (lib, ", ".join(local_filenames), retval),
                level=LOG_ALWAYS,
            )
            return retval
    if len(filenames) > 0:
        retval = filenames[0]
        if len(set(filenames)) == 1:
            return retval
        else:
            DEFAULT_LOG(
                "WARNING: Multiple non-existent physical paths match logical path %s: %s.  Selecting %s."
                % (lib, ", ".join(sorted(set(filenames))), retval),
                level=LOG_ALWAYS,
            )
            return retval
    return fix_path(
        os.path.relpath(os.path.normpath(lib.replace(".", os.sep) + ext), ".")
    )


def filename_of_lib(lib, ext=".v", **kwargs):
    kwargs = fill_kwargs(kwargs)
    return filename_of_lib_helper(
        lib,
        libnames=tuple(kwargs["libnames"]),
        non_recursive_libnames=tuple(kwargs["non_recursive_libnames"]),
        ext=ext,
    )


@memoize
def lib_of_filename_helper(filename, libnames, non_recursive_libnames, exts):
    filename = os.path.relpath(os.path.normpath(filename), ".")
    for ext in exts:
        if filename.endswith(ext):
            filename = filename[: -len(ext)]
            break
    for physical_name, logical_name in (
        (os.path.relpath(os.path.normpath(phys), "."), libname_with_dot(logical))
        for phys, logical in list(libnames) + list(non_recursive_libnames)
    ):
        filename_rel = os.path.relpath(filename, physical_name)
        if not filename_rel.startswith(".." + os.sep) and not os.path.isabs(
            filename_rel
        ):
            return (filename, logical_name + filename_rel.replace(os.sep, "."))
    if filename.startswith(".." + os.sep) and not os.path.isabs(filename):
        filename = os.path.abspath(filename)
    return (filename, filename.replace(os.sep, "."))


DEFAULT_LIB_FILENAME_EXTS = (".v", ".glob")


def lib_of_filename(filename, exts=DEFAULT_LIB_FILENAME_EXTS, **kwargs):
    kwargs = fill_kwargs(kwargs)
    filename, libname = lib_of_filename_helper(
        filename,
        libnames=tuple(kwargs["libnames"]),
        non_recursive_libnames=tuple(kwargs["non_recursive_libnames"]),
        exts=exts,
    )
    #    if '.' in filename:
    #        # TODO: Do we still need this warning?
    #        kwargs['log']("WARNING: There is a dot (.) in filename %s; the library conversion probably won't work." % filename)
    return libname


DUMMY_TOPNAME = "DUMMY TOPNAME"


def adjust_dummy_topname_binding(filename, bindings, dummy_topname=DUMMY_TOPNAME):
    if filename is None:
        return tuple(bindings)
    _, topname = lib_of_filename_helper(
        filename,
        libnames=tuple(tuple(i[1:]) for i in bindings if i[0] == "-R"),
        non_recursive_libnames=tuple(tuple(i[1:]) for i in bindings if i[0] == "-Q"),
        exts=DEFAULT_LIB_FILENAME_EXTS,
    )
    if topname.startswith("."):  # absolute path, will give other errors
        topname = os.path.splitext(os.path.basename(filename))[0]
    topname = topname.replace("-", "_DASH_")
    return tuple(
        (("-top", topname) if i == ("-top", dummy_topname) else i) for i in bindings
    )


def has_dir_binding(args, coqc_help, file_name=None):
    return any(i[0] in ("-R", "-Q") for i in group_coq_args(args, coqc_help))


def deduplicate_trailing_dir_bindings_get_topname(
    args, coqc_help, coq_accepts_top, file_name=None, topname=None
):
    kwargs = dict()
    if topname is None:
        topname = DUMMY_TOPNAME
    if file_name is not None:
        kwargs["topname"] = topname
    bindings = group_coq_args(args, coqc_help, **kwargs)
    ret = []
    for binding in adjust_dummy_topname_binding(file_name, bindings):
        if binding[0] == "-top":
            topname = binding[1]
        if coq_accepts_top or binding[0] != "-top":
            ret.extend(binding)
    return tuple(ret), topname


def deduplicate_trailing_dir_bindings(
    args, coqc_help, coq_accepts_top, file_name=None, topname=None
):
    ret, _topname = deduplicate_trailing_dir_bindings_get_topname(
        args, coqc_help, coq_accepts_top, file_name=file_name, topname=topname
    )
    return ret


def is_local_import(libname, **kwargs):
    """Returns True if libname is an import to a local file that we can discover and include, and False otherwise"""
    return os.path.isfile(filename_of_lib(libname, **kwargs))


def get_raw_file_as_bytes(filename, **kwargs):
    kwargs = fill_kwargs(kwargs)
    filename_extra = (
        "" if os.path.isabs(filename) else " (%s)" % os.path.abspath(filename)
    )
    kwargs["log"]("getting %s%s" % (filename, filename_extra))
    with open(filename, "rb") as f:
        contents = f.read()
    kwargs["log"](
        "====== Contents of %s%s ======\n%s\n==================\n"
        % (filename, filename_extra, contents.decode("utf-8")),
        level=4,
    )
    return contents


def get_raw_file(*args, **kwargs):
    return util.normalize_newlines(
        get_raw_file_as_bytes(*args, **kwargs).decode("utf-8")
    )


def list_endswith(ls, suffix):
    return len(suffix) <= len(ls) and ls[-len(suffix) :] == suffix


# code, endswith is string
@memoize
def constr_name_endswith(code, endswith):
    first_word = code.split(" ")[0].split(".")
    endswith = endswith.split(" ")[-1].split(".")
    return list_endswith(first_word, endswith)


# before, after are both strings
def move_strings_once(before, after, possibility, relaxed=False):
    for i in possibility:
        if before[-len(i) :] == i:
            return before[: -len(i)], before[-len(i) :] + after
    if relaxed:  # allow no matches
        return before, after
    else:
        return None, None


# before, after are both strings
def move_strings_pre(before, after, possibility):
    while len(before) > 0:
        new_before, new_after = move_strings_once(before, after, possibility)
        if new_before is None or new_after is None:
            return before, after
        before, after = new_before, new_after
    return (before, after)


# before, after are both strings
def move_function(before, after, get_len):
    while len(before) > 0:
        n = get_len(before)
        if n is None or n <= 0:
            return before, after
        before, after = before[:-n], before[n:] + after
    return before, after


# before, after are both strings
def move_strings(before, after, *possibilities):
    for possibility in possibilities:
        before, after = move_strings_pre(before, after, possibility)
    return before, after


# before, after are both strings
def move_space(before, after):
    return move_strings(before, after, "\n\t\r ")


# before, after are both strings
def move_comments_and_space(before, after):
    # TODO: this method is pretty inefficient, because it parses the entire string
    whitespace = "\n\t\r "
    before_without_trailing_comments_and_space = strip_trailing_comments(
        before, whitespace
    ).rstrip(whitespace)
    i = len(before_without_trailing_comments_and_space)
    return before[:i], before[i:] + after


# before, after are both strings
def move_comments_and_space_and_import_categories(before, after):
    # TODO: this method is very inefficient, because it repeatedly parses the entire prefix to move comments and space
    before, after = move_comments_and_space(before, after)
    if len(before) == 0 or before[-1] != ")":  # no import category
        return before, after
    before, after = before[:-1], before[-1] + after
    level = 1
    while level > 0:
        if len(before) == 0:
            return None, None
        before, after = move_comments_and_space(before, after)
        if len(before) == 0:
            return None, None
        if before[-1] == ")":
            level += 1
        elif before[-1] == "(":
            level -= 1
        before, after = before[:-1], before[-1] + after
    before, after = move_comments_and_space(before, after)
    # optional negation of import categories
    before, after = move_strings_once(before, after, ("-",), relaxed=True)
    before, after = move_comments_and_space(before, after)
    return before, after


# uses byte locations
def remove_from_require_before(contents, location):
    """removes "From ... " from things like "From ... Require ..." """
    assert contents is bytes(contents)
    before, after = (
        contents[:location].decode("utf-8"),
        contents[location:].decode("utf-8"),
    )
    before, after = move_comments_and_space_and_import_categories(before, after)
    if before is None or after is None:
        return contents
    before, after = move_strings_once(before, after, ("Import", "Export"), relaxed=True)
    before, after = move_comments_and_space(before, after)
    before, after = move_strings_once(before, after, ("Require",), relaxed=False)
    if before is None or after is None:
        return contents
    before, _ = move_comments_and_space(before, after)
    before, _ = move_function(
        before, after, (lambda b: 1 if b[-1] not in " \t\r\n" else None)
    )
    if before is None:
        return contents
    before, _ = move_comments_and_space(before, after)
    before, _ = move_strings_once(before, after, ("From",), relaxed=False)
    if before is None:
        return contents
    return (before + after).encode("utf-8")


# returns locations as bytes
def get_references_from_globs(globs):
    all_globs = set(
        (int(start), int(end) + 1, loc, append, ty.strip())
        for start, end, loc, append, ty in re.findall(
            "^R([0-9]+):([0-9]+) ([^ ]+) <> ([^ ]+) ([^ ]+)$", globs, flags=re.MULTILINE
        )
    )
    return tuple(sorted(all_globs, key=(lambda x: x[0]), reverse=True))


# contents should be bytes
def is_absolutizable_mod_reference(contents, reference):
    start, end, loc, append, ty = reference
    if ty != "mod":
        return False
    cur_statement = re.sub(
        r"\s+",
        " ",
        ([""] + split_coq_file_contents(contents[:start].decode("utf-8")))[-1],
    ).strip()
    return cur_statement.split(" ")[0] in ("Import", "Export", "Include")


# we absolutize requires, etc, without assuming that we have access to
# the statement-splitting of the file; this is required for inlining
# installed files, where we can't guarantee that we know how to run
# coqc on the file.
#
# contents should be bytes; globs should be string
def update_one_with_glob(
    contents, ref, absolutize, libname, transform_base=(lambda x: x), **kwargs
):
    assert contents is bytes(contents)
    kwargs = fill_kwargs(kwargs)
    start, end, loc, append, ty = ref
    cur_code = contents[start:end].decode("utf-8")
    rep = transform_base(loc) + ("." + append if append != "<>" else "")
    rep_orig = loc + ("." + append if append != "<>" else "")
    if ty not in absolutize or loc == libname:
        kwargs["log"](
            "Skipping %s at %d:%d (%s), location %s %s"
            % (ty, start, end, cur_code, loc, append),
            level=2,
        )
    # sanity check for correct replacement, to skip things like record builder notation
    elif append != "<>" and not constr_name_endswith(cur_code, append):
        kwargs["log"](
            "Skipping invalid %s at %d:%d (%s), location %s %s"
            % (ty, start, end, cur_code, loc, append),
            level=2,
        )
    elif ty == "mod" and not is_absolutizable_mod_reference(contents, ref):
        kwargs["log"](
            "Skipping non-absolutizable module reference %s at %d:%d (%s), location %s %s"
            % (ty, start, end, cur_code, loc, append),
            level=2,
        )
    elif not constr_name_endswith(rep_orig, cur_code):
        kwargs["log"](
            "Skipping invalid %s at %d:%d (%s) (expected %s), location %s %s"
            % (ty, start, end, cur_code, rep_orig, loc, append),
            level=2,
        )
    else:  # ty in absolutize and loc != libname
        kwargs["log"](
            "Qualifying %s %s to %s" % (ty, cur_code, rep), level=2, max_level=3
        )
        kwargs["log"](
            "Qualifying %s %s to %s from R%s:%s %s <> %s %s"
            % (ty, cur_code, rep, start, end, loc, append, ty),
            level=3,
        )
        contents = contents[:start] + rep.encode("utf-8") + contents[end:]
        contents = remove_from_require_before(contents, start)

    return contents


def update_with_glob(contents, globs, *args, **kwargs):
    assert contents is bytes(contents)
    for ref in get_references_from_globs(globs):
        new_contents = update_one_with_glob(contents, ref, *args, **kwargs)
        if new_contents != contents:
            last_successful_ref = ref
        contents = new_contents
    return contents


def get_all_v_files(directory, exclude=tuple()):
    all_files = []
    exclude = [os.path.normpath(i) for i in exclude]
    for dirpath, dirnames, filenames in os.walk(directory):
        all_files += [
            os.path.relpath(name, ".")
            for name in glob.glob(os.path.join(dirpath, "*.v"))
            if os.path.normpath(name) not in exclude
        ]
    return tuple(map(fix_path, all_files))


# we want to run on passing arguments if we're running in
# passing/non-passing mode, cf
# https://github.com/JasonGross/coq-tools/issues/57.  Hence we return
# the passing version iff passing_coqc is passed.
# However, if the passing coqc does not have built vo files,
# as will be the case on Coq's CI (cf https://github.com/JasonGross/coq-tools/issues/296),
# we also allow the caller to force the failing version, so that we can get the longest .glob file.
def get_maybe_passing_arg(kwargs, key, force_failing: bool = False):
    if not force_failing and kwargs.get("passing_coqc"):
        return kwargs["passing_" + key]
    return kwargs[key]


def get_keep_error_reversed(mkfile, **kwargs):
    # old versions of coq_makefile reversed the meaning of KEEP_ERROR,
    # see https://github.com/coq/coq/pull/15880.  So we need to not
    # pass it if the logic is reversed, but we should pass it
    # otherwise, because we need .glob files even if coqc fails.
    return r"""ifneq (,$(KEEP_ERROR))
.DELETE_ON_ERROR:
""" in get_raw_file(mkfile, **kwargs)


# these arguments prevent generation of .glob files
BAD_ARGS_FOR_MAKE_GLOB = ("-vio", "-vos", "-vok")


def run_coq_makefile_and_make(v_files, targets, **kwargs):
    kwargs = safe_kwargs(fill_kwargs(kwargs, for_makefile=True))
    f = tempfile.NamedTemporaryFile(
        suffix=".coq", prefix="Makefile", dir=".", delete=False
    )
    mkfile = os.path.basename(f.name)
    f.close()
    assert isinstance(kwargs["coq_makefile"], tuple), kwargs["coq_makefile"]
    assert isinstance(kwargs["coqdep"], tuple), kwargs["coqdep"]
    assert isinstance(get_maybe_passing_arg(kwargs, "coqc"), tuple), (
        get_maybe_passing_arg(kwargs, "coqc")
    )
    cmds = [
        *kwargs["coq_makefile"],
        "COQC",
        "=",
        shlex_join(get_maybe_passing_arg(kwargs, "coqc")),
        "COQDEP",
        "=",
        shlex_join(kwargs["coqdep"]),
        "-o",
        mkfile,
    ]
    for physical_name, logical_name in get_maybe_passing_arg(kwargs, "libnames"):
        cmds += [
            "-R",
            physical_name,
            (logical_name if logical_name not in ("", "''", '""') else '""'),
        ]
    for physical_name, logical_name in get_maybe_passing_arg(
        kwargs, "non_recursive_libnames"
    ):
        cmds += [
            "-Q",
            physical_name,
            (logical_name if logical_name not in ("", "''", '""') else '""'),
        ]
    for dirname in get_maybe_passing_arg(kwargs, "ocaml_dirnames"):
        cmds += ["-I", dirname]
    coq_makefile_help = get_coqc_help(kwargs["coq_makefile"], **kwargs)
    grouped_args, unrecognized_args = group_coq_args_split_recognized(
        get_maybe_passing_arg(kwargs, "coqc_args"),
        coq_makefile_help,
        is_coq_makefile=True,
    )
    for args in grouped_args:
        cmds.extend(args)
    if unrecognized_args:
        if coq_makefile_supports_arg(coq_makefile_help):
            skip_next = False
            for arg in unrecognized_args:
                # coq_makefile should never get -top{,file}; this case
                # will happen if the underlying coqc does not support
                # -top, but coqtop does, and we're running coqtop as
                # coqc
                if arg in ("-top", "-topfile"):
                    skip_next = True
                elif skip_next:
                    skip_next = False
                elif arg not in BAD_ARGS_FOR_MAKE_GLOB:
                    cmds += ["-arg", shlex_quote(arg)]
        else:
            kwargs["log"](
                "WARNING: Unrecognized arguments to coq_makefile: %s"
                % repr(unrecognized_args)
            )
    cmds += list(map(fix_path, v_files))
    kwargs["log"](shlex_join(cmds))
    try:
        p_make_makefile = subprocess.Popen(cmds, stdout=subprocess.PIPE)
        (stdout, stderr) = p_make_makefile.communicate()
    except OSError as e:
        error("When attempting to run coq_makefile:")
        error(repr(e))
        error("Failed to run coq_makefile using command line:")
        error(shlex_join(cmds))
        error("Perhaps you forgot to add COQBIN to your PATH?")
        error("Try running coqc on your files to get .glob files, to work around this.")
        sys.exit(1)
    keep_error_fragment = (
        [] if get_keep_error_reversed(mkfile, **kwargs) else ["KEEP_ERROR=1"]
    )
    make_cmds = ["make", "-k", "-f", mkfile] + keep_error_fragment + targets
    kwargs["log"](shlex_join(make_cmds))
    try:
        p_make = subprocess.Popen(
            make_cmds,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        p_make_stdout, p_make_stderr = p_make.communicate()
        return p_make_stdout.decode("utf-8")
    finally:
        for filename in (
            mkfile,
            mkfile + ".conf",
            mkfile + ".d",
            ".%s.d" % mkfile,
            ".coqdeps.d",
        ):
            if os.path.exists(filename):
                os.remove(filename)


def make_one_glob_file_helper(
    v_file, clobber_glob_on_failure: bool = True, force_failing: bool = False, **kwargs
):
    kwargs = safe_kwargs(fill_kwargs(kwargs))
    coqc_prog = get_maybe_passing_arg(kwargs, "coqc", force_failing=force_failing)
    coqpath_paths = get_maybe_passing_arg(
        kwargs, "coqpath_paths", force_failing=force_failing
    )
    assert isinstance(coqc_prog, tuple), coqc_prog
    cmds = [*coqc_prog, "-q"]
    for physical_name, logical_name in get_maybe_passing_arg(
        kwargs, "libnames", force_failing=force_failing
    ):
        cmds += [
            "-R",
            physical_name,
            (logical_name if logical_name not in ("", "''", '""') else '""'),
        ]
    for physical_name, logical_name in get_maybe_passing_arg(
        kwargs, "non_recursive_libnames", force_failing=force_failing
    ):
        cmds += [
            "-Q",
            physical_name,
            (logical_name if logical_name not in ("", "''", '""') else '""'),
        ]
    for dirname in get_maybe_passing_arg(
        kwargs, "ocaml_dirnames", force_failing=force_failing
    ):
        cmds += ["-I", dirname]
    cmds += list(
        get_maybe_passing_arg(kwargs, "coqc_args", force_failing=force_failing)
    )
    v_file = fix_path(v_file)
    v_file_root, ext = os.path.splitext(v_file)
    o_file = os.path.join(tempfile.gettempdir(), os.path.basename(v_file_root) + ".vo")
    tmp_glob_file = os.path.join(
        tempfile.gettempdir(), os.path.basename(v_file_root) + ".glob"
    )
    glob_file = v_file_root + ".glob"
    if get_coq_accepts_o(coqc_prog, **kwargs):
        cmds += ["-o", o_file]
    else:
        kwargs["log"](
            "WARNING: Clobbering '%s' because coqc does not support -o" % o_file,
            level=LOG_ALWAYS,
        )
    cmds += ["-dump-glob", tmp_glob_file, v_file]
    # if the file has user-contrib in it, we may try to its local paths with -R for more robustness
    extra_cmds_options = []
    v_file_in_coqpaths = [p for p in coqpath_paths if is_subdirectory_of(v_file, p)]
    # sort the paths by length of absolute path so that we try the longest paths first
    v_file_in_coqpaths.sort(key=lambda p: len(os.path.abspath(p)), reverse=True)
    for coqpath_path in v_file_in_coqpaths:
        relative_path = os.path.relpath(v_file, coqpath_path)
        first_component = fix_path(relative_path).split("/")[0]
        extra_cmds_options.append(
            ["-R", os.path.join(coqpath_path, first_component), first_component]
        )
    kwargs["log"](shlex_join(cmds))
    try:
        p = subprocess.Popen(cmds, stdout=subprocess.PIPE)
        stdout, stderr = p.communicate()
        for extra_cmds in extra_cmds_options:
            if p.returncode == 0:
                break
            kwargs["log"](
                "WARNING: coqc failed on '%s', retrying with extra %s"
                % (v_file, shlex_join(extra_cmds)),
                level=LOG_ALWAYS,
            )
            kwargs["log"](shlex_join(cmds + extra_cmds))
            p = subprocess.Popen(cmds + extra_cmds, stdout=subprocess.PIPE)
            stdout, stderr = p.communicate()
        if os.path.exists(tmp_glob_file) and (
            p.returncode == 0
            or not os.path.exists(glob_file)
            or clobber_glob_on_failure
        ):
            if os.path.exists(glob_file):
                extra = (
                    " despite failure of coqc"
                    if p.returncode != 0
                    else " because coqc succeeded"
                )
                kwargs["log"](
                    f"WARNING: Clobbering '{glob_file}' ({os.path.getmtime(glob_file)}) from '{v_file}' ({os.path.getmtime(v_file_root + ext)}){extra}",
                    level=LOG_ALWAYS,
                )
            else:
                kwargs["log"](f"Moving '{tmp_glob_file}' to '{glob_file}'")
            try:
                # N.B. We need shutil.move rather than os.rename because the source and destination may be on different filesystems
                shutil.move(tmp_glob_file, glob_file)
            except PermissionError as e:
                kwargs["log"](
                    f"Failed to move '{tmp_glob_file}' to '{glob_file}' ({e}), assuming that '{glob_file}' is up to date"
                )
        elif os.path.exists(tmp_glob_file):
            kwargs["log"](
                f"WARNING: Not clobbering '{glob_file}' ({os.path.getmtime(glob_file)}) because coqc failed on '{v_file}' ({os.path.getmtime(v_file_root + ext)})",
                level=LOG_ALWAYS,
            )
        return stdout, stderr
    finally:
        if os.path.exists(o_file):
            os.remove(o_file)
        if os.path.exists(tmp_glob_file):
            os.remove(tmp_glob_file)


# we want to run on passing arguments if we're running in
# passing/non-passing mode, cf
# https://github.com/JasonGross/coq-tools/issues/57.  Hence we return
# the passing version iff passing_coqc is passed.
# However, if the passing coqc does not have built vo files,
# as will be the case on Coq's CI (cf https://github.com/JasonGross/coq-tools/issues/296),
# we also allow the caller to force the failing version, so that we can get the longest .glob file.
def get_glob_contents(v_file, **kwargs):
    kwargs = safe_kwargs(fill_kwargs(kwargs))
    v_file = Path(fix_path(v_file))
    glob_file = v_file.with_suffix(".glob")
    stdout, stderr = make_one_glob_file_helper(str(v_file), **kwargs)
    return (stdout, stderr), glob_file.read_text()


def make_one_glob_file(v_file, **kwargs):
    kwargs = safe_kwargs(fill_kwargs(kwargs))
    v_file = Path(fix_path(v_file))
    glob_file = v_file.with_suffix(".glob")
    (stdout_failing, stderr_failing), glob_contents_failing = get_glob_contents(
        str(v_file), force_failing=True, **kwargs
    )
    if not kwargs.get("passing_coqc"):
        return (stdout_failing, stderr_failing), (None, None)
    (stdout_passing, stderr_passing), glob_contents_passing = get_glob_contents(
        str(v_file), force_failing=False, **kwargs
    )
    if glob_contents_failing != glob_contents_passing:
        passing_lines = len(glob_contents_passing.splitlines())
        failing_lines = len(glob_contents_failing.splitlines())
        kwargs["log"](f"Failing .glob file:\n\n{glob_contents_failing}\n\n", level=4)
        kwargs["log"](f"Passing .glob file:\n\n{glob_contents_passing}\n\n", level=4)
        if failing_lines > passing_lines:
            kwargs["log"](
                f"Taking the failing {glob_file} file ({failing_lines} lines) because the passing version ({passing_lines} lines) is shorter"
            )
            glob_file.write_text(glob_contents_failing)
        else:
            kwargs["log"](
                f"Taking the passing {glob_file} file ({passing_lines} lines) because the failing version ({failing_lines} lines) is longer"
            )
            glob_file.write_text(glob_contents_passing)
    return (stdout_failing, stderr_failing), (stdout_passing, stderr_passing)


def is_subdirectory_of(filename, parent):
    abs_filename = os.path.normpath(os.path.abspath(filename))
    parent = os.path.normpath(os.path.abspath(parent))
    common = os.path.normpath(os.path.commonprefix([parent, abs_filename]))
    return common == parent


def is_local(filename):
    return is_subdirectory_of(filename, ".")


def remove_if_local(filename, **kwargs):
    abs_filename = os.path.abspath(filename)
    cwd = os.path.abspath(".")
    common = os.path.commonprefix([cwd, abs_filename])
    if common != cwd:
        kwargs["log"](
            "WARNING: Not removing %s (%s) because it resides in a parent (%s) of the current directory (%s)"
            % (filename, abs_filename, common, cwd)
        )
        return
    os.remove(filename)


def make_globs(logical_names, **kwargs):
    kwargs = fill_kwargs(kwargs)
    existing_logical_names = [
        i
        for i in logical_names
        if os.path.isfile(filename_of_lib(i, ext=".v", **kwargs))
    ]
    if len(existing_logical_names) == 0:
        return
    filenames_vo_v_glob = [
        (
            filename_of_lib(i, ext=".vo", **kwargs),
            filename_of_lib(i, ext=".v", **kwargs),
            filename_of_lib(i, ext=".glob", **kwargs),
        )
        for i in existing_logical_names
    ]
    filenames_vo_v_glob = [
        (vo_name, v_name, glob_name)
        for vo_name, v_name, glob_name in filenames_vo_v_glob
        if not (
            os.path.isfile(glob_name)
            and os.path.getmtime(glob_name) > os.path.getmtime(v_name)
        )
    ]
    for vo_name, v_name, glob_name in filenames_vo_v_glob:
        if os.path.isfile(glob_name) and not os.path.getmtime(
            glob_name
        ) > os.path.getmtime(v_name):
            if os.path.getmtime(v_name) > time.time():
                kwargs["log"](
                    "WARNING: The file %s comes from the future! (%d > %d)"
                    % (v_name, os.path.getmtime(v_name), time.time()),
                    level=LOG_ALWAYS,
                )
            remove_if_local(glob_name, **kwargs)
        # if the .vo file already exists and is new enough, we assume
        # that all dependent .vo files also exist, and just run coqc
        # in a way that doesn't update the .vo file.  We use >= rather
        # than > because we're using .vo being new enough as a proxy
        # for the dependent .vo files existing, so we don't care as
        # much about being perfectly accurate on .vo file timing
        # (unlike .glob file timing, where we need it to be up to
        # date), and it's better to not clobber the .vo file when
        # we're unsure if it's new enough.
        ambiguous_glob_time = os.path.isfile(glob_name) and os.path.getmtime(
            glob_name
        ) == os.path.getmtime(v_name)
        # We also don't want to clobber the .glob file from ambiguous non-local (e.g., installed) files if coqc fails
        if os.path.exists(vo_name) and os.path.getmtime(vo_name) >= os.path.getmtime(
            v_name
        ):
            make_one_glob_file(
                v_name,
                clobber_glob_on_failure=is_local(glob_name) or not ambiguous_glob_time,
                **kwargs,
            )
    filenames_vo_v_glob = [
        (vo_name, v_name, glob_name)
        for vo_name, v_name, glob_name in filenames_vo_v_glob
        if not (
            os.path.exists(vo_name)
            and os.path.getmtime(vo_name) >= os.path.getmtime(v_name)
        )
    ]
    filenames_v = [v_name for vo_name, v_name, glob_name in filenames_vo_v_glob]
    filenames_glob = [glob_name for vo_name, v_name, glob_name in filenames_vo_v_glob]
    if len(filenames_vo_v_glob) == 0:
        return
    extra_filenames_v = list(
        get_all_v_files(".", filenames_v)
        if kwargs["use_coq_makefile_for_deps"] and kwargs["walk_tree"]
        else []
    )
    all_filenames_v = filenames_v + extra_filenames_v

    if len(all_filenames_v) <= 1 or not kwargs["use_coq_makefile_for_deps"]:
        if len(all_filenames_v) > 1:
            kwargs["log"](
                "WARNING: Making globs for %s separately without coq_makefile for dependencies, as requested"
                % ", ".join(all_filenames_v)
            )
        for v_name in all_filenames_v:
            v_name_base, _ext = os.path.splitext(v_name)
            glob_name = v_name_base + ".glob"
            ambiguous_glob_time = os.path.isfile(glob_name) and os.path.getmtime(
                glob_name
            ) == os.path.getmtime(v_name)
            make_one_glob_file(
                v_name,
                clobber_glob_on_failure=is_local(glob_name) and not ambiguous_glob_time,
                **kwargs,
            )
    else:
        stdouterr_make = run_coq_makefile_and_make(
            tuple(sorted(filenames_v + extra_filenames_v)), filenames_glob, **kwargs
        )
        if "Argument list too long" in stdouterr_make and re.search(
            r"No rule to make target '.*\.glob'", stdouterr_make
        ):
            kwargs["log"](
                "WARNING: Making globs may not have succeeded, consider passing "
                + kwargs["cli_mapping"]["use_coq_makefile_for_deps"][False][0]
            )


def get_glob_file_for(
    filename, update_globs=False, allow_ambiguous_glob_files: bool = True, **kwargs
):
    kwargs = fill_kwargs(kwargs)
    filename = fix_path(filename)
    if filename[-2:] != ".v":
        filename += ".v"
    libname = lib_of_filename(filename, **kwargs)
    globname = filename[:-2] + ".glob"
    if not os.path.exists(filename):
        return None
    if (
        filename not in raw_file_contents.keys()
        or file_mtimes[filename] < os.stat(filename).st_mtime
    ):
        raw_file_contents[filename] = get_raw_file_as_bytes(filename, **kwargs)
        file_contents[filename] = {}
        file_mtimes[filename] = os.stat(filename).st_mtime
    if update_globs:
        if file_mtimes[filename] > time.time():
            kwargs["log"](
                "WARNING: The file %s comes from the future! (%d > %d)"
                % (filename, file_mtimes[filename], time.time()),
                level=LOG_ALWAYS,
            )
        if time.time() - file_mtimes[filename] < 2:
            kwargs["log"](
                "NOTE: The file %s is very new (%d, %d seconds old), delaying until it's a bit older"
                % (filename, file_mtimes[filename], time.time() - file_mtimes[filename])
            )
        # delay until the .v file is old enough that a .glob file will be considered newer
        # if we just wait until they're not equal, we apparently get issues like https://gitlab.com/Zimmi48/coq/-/jobs/535005442
        while time.time() - file_mtimes[filename] < 2:
            time.sleep(0.1)
        make_globs([libname], **kwargs)
    if os.path.isfile(globname):
        if os.stat(globname).st_mtime > file_mtimes[filename]:
            return get_raw_file(globname, **kwargs)
        elif (
            allow_ambiguous_glob_files
            and os.stat(globname).st_mtime == file_mtimes[filename]
        ):
            kwargs["log"](
                "WARNING: Ambiguous .glob file %s for %s (%d)"
                % (globname, filename, file_mtimes[filename])
            )
            return get_raw_file(globname, **kwargs)
        else:
            kwargs["log"](
                "WARNING: Assuming that %s is not a valid reflection of %s because %s is newer (%d >= %d)"
                % (
                    globname,
                    filename,
                    filename,
                    file_mtimes[filename],
                    os.stat(globname).st_mtime,
                )
            )
    return None


def get_byte_references_for(filename, types=None, appends=None, **kwargs):
    globs = get_glob_file_for(filename, **kwargs)
    if globs is None:
        return None
    references = get_references_from_globs(globs)
    return tuple(
        (start, end, loc, append, ty)
        for start, end, loc, append, ty in references
        if (types is None or ty in types) and (appends is None or append in appends)
    )


def get_file_as_bytes(
    filename,
    absolutize=("lib",),
    update_globs=False,
    mod_remap=dict(),
    transform_base=(lambda x: x),
    **kwargs,
):
    kwargs = fill_kwargs(kwargs)
    filename = fix_path(filename)
    if filename[-2:] != ".v":
        filename += ".v"
    libname = lib_of_filename(filename, **kwargs)
    globname = filename[:-2] + ".glob"
    if (
        filename not in raw_file_contents.keys()
        or file_mtimes[filename] < os.stat(filename).st_mtime
    ):
        raw_file_contents[filename] = get_raw_file_as_bytes(filename, **kwargs)
        file_contents[filename] = {}
        file_mtimes[filename] = os.stat(filename).st_mtime
    if len(mod_remap.keys()) > 0:
        absolutize = list(set(list(absolutize) + ["mod"]))
    key = (tuple(sorted(absolutize)), tuple(sorted(list(mod_remap.items()))))
    if key not in file_contents[filename].keys():
        if len(absolutize) > 0 or len(mod_remap.keys()) > 0:
            globs = get_glob_file_for(filename, update_globs=update_globs, **kwargs)
            if globs is not None:
                file_contents[filename][key] = update_with_glob(
                    raw_file_contents[filename],
                    globs,
                    absolutize,
                    libname,
                    transform_base=(lambda x: mod_remap.get(x, transform_base(x))),
                    **kwargs,
                )
    return file_contents[filename].get(key, raw_file_contents[filename])


# returns string, newlines normalized
def get_file(*args, **kwargs):
    return util.normalize_newlines(get_file_as_bytes(*args, **kwargs).decode("utf-8"))


# this is used to mark which statements are tied to which requires
def insert_references(
    contents, ranges, references, types=("lib",), appends=("<>",), **kwargs
):
    assert contents is bytes(contents)
    ret = []
    prev = 0
    bad = [
        (start, end, loc, append, ty)
        for start, end, loc, append, ty in references
        if append not in appends or ty not in types
    ]
    if bad:
        kwargs["log"](
            "Invalid glob entries: %s"
            % "\n".join(
                "R%d:%d %s <> %s %s" % (start, end, loc, append, ty)
                for start, end, loc, append, ty in bad
            )
        )
    for start, finish in ranges:
        if prev < start:
            ret.append((contents[prev:start], tuple()))
        bad = [
            (rstart, rend, loc, append, ty)
            for rstart, rend, loc, append, ty in references
            if (start <= rstart or rend <= finish)
            and not (
                (start <= rstart and rend <= finish)
                or finish <= rstart
                or rend <= start
            )
        ]
        if bad:
            kwargs["log"](
                "Invalid glob entries partially overlapping (%d, %d]: %s"
                % (
                    start,
                    finish,
                    "\n".join(
                        "R%d:%d %s <> %s %s" % (start, end, loc, append, ty)
                        for start, end, loc, append, ty in bad
                    ),
                )
            )

        cur_references = tuple(
            (rstart - start, rend - start, loc, append, ty)
            for rstart, rend, loc, append, ty in references
            if start <= rstart and rend <= finish
        )
        ret.append((contents[start:finish], cur_references))
        prev = finish
    if prev < len(contents):
        ret.append((contents[prev:], tuple()))
    return tuple(ret)


def get_file_statements_insert_references(filename, **kwargs):
    ranges = get_coq_statement_byte_ranges(filename, **kwargs)
    contents = get_file_as_bytes(filename, **kwargs)
    refs = get_byte_references_for(filename, **kwargs)
    if refs is None:
        return None
    return insert_references(contents, ranges, refs, **kwargs)


def remove_after_first_range(text_as_bytes, ranges):
    if ranges:
        start = min((r[0] for r in ranges))
        return text_as_bytes[:start]
    else:
        return text_as_bytes


REQUIRE, REQUIRE_IMPORT, REQUIRE_EXPORT, EXPORT, IMPORT = (
    "REQUIRE",
    "REQUIRE IMPORT",
    "REQUIRE EXPORT",
    "EXPORT",
    "IMPORT",
)


def classify_require_kind(text_as_bytes, ranges):
    """cassifies the kind of require statement

    >>> res = {REQUIRE:'R', REQUIRE_IMPORT:'RI', REQUIRE_EXPORT:'RE', EXPORT:'E', IMPORT:'I', None:None}
    >>> print(res[classify_require_kind(b'From Coq Require Export ZArith.ZArith.', ((24, 37, 'Coq.ZArith.ZArith'),))])
    RE
    """
    prefix = (
        strip_comments(
            re.sub(
                r"\s+",
                " ",
                remove_after_first_range(text_as_bytes, ranges).decode("utf-8"),
            )
        ).strip()
        + " "
    )
    is_require = "Require " in prefix
    if "Import " in prefix:
        return REQUIRE_IMPORT if is_require else IMPORT
    if "Export " in prefix:
        return REQUIRE_EXPORT if is_require else EXPORT
    return REQUIRE if is_require else None


def split_requires_of_statements(
    annotated_contents, types=("lib",), appends=("<>",), **kwargs
):
    already_required = set()
    prefix, postfix = "\nRequire ", "."
    for statement, references in annotated_contents:
        statement_str = statement.decode("utf-8")
        if not references:
            yield (statement, references)
            continue
        rkind = classify_require_kind(statement, references)
        if rkind not in (REQUIRE, REQUIRE_IMPORT, REQUIRE_EXPORT) or (
            len(references) == 1 and rkind == REQUIRE
        ):
            yield (statement, references)
            continue
        remaining_references = []
        emitted_require = False
        for start, end, loc, append, ty in references:
            if ty in types and append in appends:
                if loc not in already_required:
                    already_required.add(loc)
                    emitted_require = True
                    yield (
                        (prefix + loc + postfix).encode("utf-8"),
                        (len(prefix), len(prefix) - start + end, loc, append, ty),
                    )
            else:
                remaining_references.append(
                    (start - len("Require "), end - len("Require "), loc, append, ty)
                )
        if rkind == REQUIRE:
            continue
        statement_str = re.sub(r"Require\s", "", statement_str, count=1)
        if emitted_require and not re.match(r"\s", statement_str[0]):
            yield ("\n".encode("utf-8"), tuple())  # we need an extra newline
        yield (statement_str.encode("utf-8"), tuple(remaining_references))


def get_require_dict(lib, update_globs=True, **kwargs):
    kwargs = fill_kwargs(kwargs)
    lib = norm_libname(lib, **kwargs)
    v_name = filename_of_lib(lib, ext=".v", **kwargs)
    if lib not in lib_imports_slow.keys():
        globs = get_byte_references_for(
            v_name, types=("lib",), appends=("<>",), update_globs=update_globs, **kwargs
        )
        if globs is not None:
            lib_imports_slow[lib] = {}
            for start, end, name, append, ty in reversed(globs):
                name = norm_libname(name, **kwargs)
                if name not in lib_imports_slow[lib].keys():
                    lib_imports_slow[lib][name] = []
                lib_imports_slow[lib][name].append((start, end))
            for name in lib_imports_slow[lib].keys():
                lib_imports_slow[lib][name] = tuple(lib_imports_slow[lib][name])
    return lib_imports_slow.get(lib, {})


def get_require_names(lib, **kwargs):
    return tuple(sorted(get_require_dict(lib, **kwargs).keys()))


def get_require_locations(lib, **kwargs):
    return sorted(
        set(
            loc
            for name, locs in get_require_dict(lib, **kwargs).items()
            for loc in locs
        )
    )


def transitively_close(d, make_new_value=(lambda x: tuple()), reflexive=True):
    updated = True
    while updated:
        updated = False
        for key in tuple(d.keys()):
            newv = set(d[key])
            if reflexive:
                newv.add(key)
            for v in tuple(newv):
                if v not in d.keys():
                    d[v] = make_new_value(v)
                newv.update(set(d[v]))
            if newv != set(d[key]):
                d[key] = newv
                updated = True
    return d


@memoize
def compare_lib_prefixes(prefix1, prefix2, requires_dict_items):
    requires_dict = dict(requires_dict_items)
    all_libs = list(requires_dict.keys())
    assert len(prefix1) == len(prefix2), (prefix1, prefix2)
    first_diff_idx = len(prefix1)
    any_d1_prime_in_d2_prime = False
    any_d2_prime_in_d1_prime = False
    for d1_prime in all_libs:
        d1_prime_parts = d1_prime.split(".")
        if (
            len(d1_prime_parts) > first_diff_idx
            and d1_prime_parts[:first_diff_idx] == prefix1
        ):
            for d2_prime in all_libs:
                d2_prime_parts = d2_prime.split(".")
                if (
                    len(d2_prime_parts) > first_diff_idx
                    and d2_prime_parts[:first_diff_idx] == prefix2
                ):
                    if d1_prime in requires_dict[d2_prime]:
                        any_d1_prime_in_d2_prime = True
                    elif d2_prime in requires_dict[d1_prime]:
                        any_d2_prime_in_d1_prime = True
    if any_d1_prime_in_d2_prime and not any_d2_prime_in_d1_prime:
        return -1
    elif any_d2_prime_in_d1_prime and not any_d1_prime_in_d2_prime:
        return 1
    else:
        return 0


def compare_libs(
    d1,
    d2,
    *,
    requires_dict: dict,
    sort_requires_by_component: bool = False,
    requires_dict_items: Optional[Tuple[Tuple[str, Tuple[str, ...]], ...]] = None,
    all_libs=None,
    fallback_cmp=None,
):
    if requires_dict_items is None:
        requires_dict_items = tuple(
            sorted((k, tuple(sorted(v))) for k, v in requires_dict.items())
        )
    if all_libs is None:
        all_libs = list(requires_dict.keys())
    if d1 == d2:
        return 0

    # Primary ordering: d1 comes before d2 if d1 in requires_dict[d2]
    if d1 in requires_dict[d2]:
        return -1
    if d2 in requires_dict[d1]:
        return 1

    # Intermediate tie-breaker: component-based comparison if sort_requires_by_component
    if sort_requires_by_component:
        d1_parts = d1.split(".")
        d2_parts = d2.split(".")

        # Find first non-equal component
        min_len = min(len(d1_parts), len(d2_parts))
        first_diff_idx = None
        for i in range(min_len):
            if d1_parts[i] != d2_parts[i]:
                first_diff_idx = i
                break

        if first_diff_idx is not None:
            # Check if there exist d1', d2' where:
            # - d1.split(".")[i] == d1'.split(".")[i] for all i < first_diff_idx
            # - d2.split(".")[i] == d2'.split(".")[i] for all i < first_diff_idx
            # - d1' in recur_dep_graph[d2']
            prefix1 = d1_parts[:first_diff_idx]
            prefix2 = d2_parts[:first_diff_idx]

            lib_cmp = compare_lib_prefixes(prefix1, prefix2, requires_dict_items)
            if lib_cmp != 0:
                return lib_cmp

    # Final tie-breaker: lexicographic comparison of split components
    if isinstance(fallback_cmp, str) and fallback_cmp.lower() == "lexicographic":
        d1_parts = d1.split(".")
        d2_parts = d2.split(".")
        if d1_parts < d2_parts:
            return -1
        elif d1_parts > d2_parts:
            return 1
        else:
            return 0
    elif fallback_cmp is not None and isinstance(fallback_cmp, Callable):
        return fallback_cmp(d1, d2)
    else:
        return 0


def get_recursive_requires(
    *libnames, sort_requires_by_component: bool = False, **kwargs
):
    reverse = kwargs.get("reverse", True)
    requires = dict((lib, get_require_names(lib, **kwargs)) for lib in libnames)
    transitively_close(
        requires,
        make_new_value=(lambda lib: get_require_names(lib, **kwargs)),
        reflexive=True,
    )

    all_libs = list(requires.keys())
    requires_dict_items = tuple(
        sorted((k, tuple(sorted(v))) for k, v in requires.items())
    )

    def cmp_by_len(l1, l2):
        # this only works correctly if the closure is *reflexive* as
        # well as transitive, because we require that if A requires B,
        # then A must have strictly more requires than B (i.e., it
        # must include itself)
        if len(requires[l1]) != len(requires[l2]):
            return cmp(len(requires[l1]), len(requires[l2]))
        return cmp(l1, l2)

    def lcmp(l1, l2):
        l1, l2 = l1[0], l2[0]  # strip off value part
        if l1 == l2:
            return cmp(l1, l2)

        return compare_libs(
            l1,
            l2,
            requires_dict=requires,
            all_libs=all_libs,
            requires_dict_items=requires_dict_items,
            sort_requires_by_component=sort_requires_by_component,
            fallback_cmp=cmp_by_len,
        )

    return OrderedDict(sorted(requires.items(), key=cmp_to_key(lcmp), reverse=reverse))


def get_recursive_require_names(libname, **kwargs):
    requires = get_recursive_requires(libname, **kwargs)
    return tuple(i for i in requires.keys() if i != libname)


def sort_files_by_dependency(filenames, reverse=True, **kwargs):
    kwargs = fill_kwargs(kwargs)
    filenames = map(fix_path, filenames)
    filenames = [
        (filename + ".v" if filename[-2:] != ".v" else filename)
        for filename in filenames
    ]
    libname_map = dict(
        (lib_of_filename(filename, **kwargs), filename) for filename in filenames
    )
    requires = get_recursive_requires(
        *sorted(libname_map.keys()), reverse=reverse, **kwargs
    )
    # filter the sorted requires by the ones that were in the original
    # filenames list, and use libname_map to lookup the original
    # filename.  We could probably just map filename_of_lib over the
    # requires and keep the ones that are in the original filenames,
    # but this is more robust if there are edge cases where
    # filename_of_lib(lib_of_filename(x)) != x
    filenames = [
        libname_map[libname]
        for libname in requires.keys()
        if libname in libname_map.keys()
    ]
    return filenames


def get_imports(lib, fast=False, **kwargs) -> Tuple[str, ...]:
    kwargs = fill_kwargs(kwargs)
    lib = norm_libname(lib, **kwargs)
    glob_name = filename_of_lib(lib, ext=".glob", **kwargs)
    v_name = filename_of_lib(lib, ext=".v", **kwargs)
    if not fast:
        get_require_dict(lib, **kwargs)
        if lib in lib_imports_slow.keys():
            return tuple(
                k
                for k, v in sorted(
                    lib_imports_slow[lib].items(), key=(lambda kv: kv[1])
                )
            )
    # making globs failed, or we want the fast way, fall back to regexp
    if lib not in lib_imports_fast.keys():
        contents = get_file(v_name, **kwargs)
        imports_string = re.sub(
            "\\s+", " ", " ".join(IMPORT_LINE_REG.findall(contents))
        ).strip()
        lib_imports_fast[lib] = tuple(
            sorted(
                set(
                    norm_libname(i, **kwargs)
                    for i in imports_string.split(" ")
                    if i != ""
                )
            )
        )
    return lib_imports_fast[lib]


def norm_libname(lib, **kwargs):
    kwargs = fill_kwargs(kwargs)
    filename = filename_of_lib(lib, **kwargs)
    if os.path.isfile(filename):
        return lib_of_filename(filename, **kwargs)
    else:
        return lib


def merge_imports(imports, **kwargs):
    kwargs = fill_kwargs(kwargs)
    rtn = []
    for import_list in imports:
        for i in import_list:
            if norm_libname(i, **kwargs) not in rtn:
                rtn.append(norm_libname(i, **kwargs))
    return rtn


# This is a bottleneck for more than around 10,000 lines of code total with many imports (around 100)
@memoize
def internal_recursively_get_imports(lib, **kwargs):
    return run_recursively_get_imports(
        lib, recur=internal_recursively_get_imports, **kwargs
    )


def recursively_get_imports(lib, **kwargs):
    return internal_recursively_get_imports(lib, **safe_kwargs(kwargs))


def run_recursively_get_imports(
    lib, recur=recursively_get_imports, fast=False, **kwargs
):
    kwargs = fill_kwargs(kwargs)
    lib = norm_libname(lib, **kwargs)
    glob_name = filename_of_lib(lib, ext=".glob", **kwargs)
    v_name = filename_of_lib(lib, ext=".v", **kwargs)
    if os.path.isfile(v_name):
        imports = get_imports(lib, fast=fast, **kwargs)
        if not fast:
            make_globs(imports, **kwargs)
        imports_list = [recur(k, fast=fast, **kwargs) for k in imports]
        return merge_imports(tuple(map(tuple, imports_list + [[lib]])), **kwargs)
    return [lib]


def run_maybe_recursively_get_imports(lib, recursively: bool = True, **kwargs):
    recursive_imports = run_recursively_get_imports(lib, **kwargs)
    if recursively:
        return recursive_imports
    else:
        # keep the order of the recursive version, but only keep the elements in the non-recursive list
        direct_imports = run_recursively_get_imports(
            lib, recur=lambda lib, **_kwargs: [lib], **kwargs
        )
        return [lib for lib in recursive_imports if lib in direct_imports]


class DepGraph(OrderedDict, Dict[str, Tuple[str, ...]]):
    def __init__(self, *args, fast=False, **kwargs):
        super().__init__(*args)
        self._fast = fast
        self._kwargs = fill_kwargs(kwargs)

    def __missing__(self, key):
        lib = norm_libname(key, **self._kwargs)
        if key != lib:
            return self[lib]
        v_name = filename_of_lib(lib, ext=".v", **self._kwargs)
        self[lib] = ()
        if os.path.isfile(v_name):
            imports = get_imports(lib, fast=self._fast, **self._kwargs)
            if not self._fast:
                make_globs(imports, **self._kwargs)
            self[lib] = imports

        return self[key]


def recursively_get_imports_(
    lib,
    sort_requires_by_component: bool = True,
    fast: bool = False,
    recursively: bool = True,
    **kwargs,
):
    dep_graph = DepGraph(fast=fast, **kwargs)
    direct_imports = dep_graph[lib]
    recur_dep_graph = transitive_closure(dep_graph)
    # Create sorted list of recursive imports
    all_libs = list(recur_dep_graph.keys())

    recursive_imports = sorted(
        all_libs,
        key=cmp_to_key(
            partial(
                compare_libs,
                requires_dict=recur_dep_graph,
                all_libs=all_libs,
                sort_requires_by_component=sort_requires_by_component,
            )
        ),
    )

    if recursively:
        return recursive_imports
    else:
        # keep the order of the recursive version, but only keep the elements in the non-recursive list
        return [lib for lib in recursive_imports if lib in direct_imports]


if __name__ == "__main__":
    # if we're working in Python 3.3, we can test this file
    try:
        import doctest

        success = True
    except ImportError:
        print(
            "This is not the main file to use.\nOnly run it if you have doctest (Python 3.3+) and are testing things."
        )
        success = False
    if success:
        doctest.testmod()



from .custom_arguments import DEFAULT_LOG

__all__ = ["process_maybe_list"]


def process_maybe_list(ls, log=DEFAULT_LOG):
    if ls is None:
        return tuple()
    if isinstance(ls, str):
        return tuple([ls])
    if isinstance(ls, tuple):
        return ls
    if isinstance(ls, list):
        return tuple(ls)
    log("Unknown type '%s' of list '%s'" % (str(type(ls)), repr(ls)))
    return tuple(ls)



import re

__all__ = ["transform_abstract_to_admit"]


def DEFAULT_LOG(text, level=1):
    if level <= 1:
        print(text)


TERM_CHAR = r"[\w']"
ABSTRACT_NO_PARENS_DOT = re.compile(
    r"(^\s*|\.\s+|;\s*)abstract\s+(?:[^\(\);\.]|\.%s)+(?=\.\s|\.$)" % TERM_CHAR,
    re.MULTILINE,
)


def transform_abstract_to_admit_statement(statement, aggressive=False, log=DEFAULT_LOG):
    # remove the unparenthesized ones
    statement = ABSTRACT_NO_PARENS_DOT.sub(r"\1admit", statement)

    if "abstract" not in statement:
        return statement

    # now look at the parenthesized abstracts
    ready_for_abstract = True
    in_abstract = False
    abstract_paren_level = 0
    rtn = []
    cur = []
    for term in re.split(r"([;\.\(\)])", statement):
        log(
            "in_abstract: %d; abstract_paren_level: %d; aggressive: %d; ready_for_abstract: %d;\n^term: %s"
            % (in_abstract, abstract_paren_level, aggressive, ready_for_abstract, term),
            level=3,
        )
        if in_abstract:
            if abstract_paren_level == 0 and term in tuple(";."):
                if term == ";":
                    if aggressive:
                        rtn.append(" admit;")
                        log("Appending ' admit;' to rtn", level=3)
                    else:
                        cur.append(term)
                        rtn.append("".join(cur))
                        log("Appending %s to rtn" % repr("".join(cur)), level=3)
                else:
                    log("Appending ' admit.' to rtn", level=3)
                    rtn.append(" admit.")
                in_abstract = False
                cur = []
                log("Setting in_abstract to false and emptying cur", level=3)
            elif abstract_paren_level < 0:
                log(
                    "Warning: abstract_paren_level messed up on statement %s"
                    % repr(statement)
                )
                in_abstract = False
                cur.append(term)
                rtn.append("".join(cur))
                cur = []
                abstract_paren_level = 0
                log(
                    "Setting in_abstract to false, abstract_paren_level to 0, emptying cur, and\nappending %s to rtn"
                    % repr("".join(cur)),
                    level=3,
                )
            else:
                if term == "(":
                    abstract_paren_level += 1
                elif term == ")":
                    abstract_paren_level -= 1
                cur.append(term)
                log(
                    "Setting abstract_paren_level to %d and\nappending %s to cur"
                    % (abstract_paren_level, repr(term)),
                    level=3,
                )
        else:
            if ready_for_abstract and term.strip() == "abstract":
                cur.append(term)
                in_abstract = True
                log(
                    "Found %s (appending to cur), now in_abstract" % repr(term), level=3
                )
                if abstract_paren_level != 0:
                    log(
                        "Warning: abstract_paren_level messed up before abstract on statement %s"
                        % repr(statement)
                    )
                    abstract_paren_level = 0
            else:
                if term in tuple(".;"):
                    log(
                        "Found %s (appending to rtn), ready for abstract" % repr(term),
                        level=3,
                    )
                    ready_for_abstract = True
                elif term.strip():
                    log(
                        "Found %s (appending to rtn), not ready for abstract"
                        % repr(term),
                        level=3,
                    )
                    ready_for_abstract = False
                rtn.append(term)
    log("Done.  Appending %s to rtn." % repr("".join(cur)), level=3)
    rtn.append("".join(cur))

    return "".join(rtn)


def transform_abstract_to_admit(cur_definition, rest_definitions, **kwargs):
    # shallow copy
    cur_definition = dict(cur_definition)
    cur_definition["statements"] = tuple(
        transform_abstract_to_admit_statement(i, **kwargs)
        for i in cur_definition["statements"]
    )
    cur_definition["statement"] = "\n".join(cur_definition["statements"])
    return cur_definition



import re
from subprocess import Popen, PIPE, STDOUT
from . import split_definitions_old
from .split_file import postprocess_split_proof_term
from .coq_version import get_coq_accepts_time, get_coq_accepts_emacs
from .coq_running_support import get_proof_term_works_with_time, parse_options_settings
from .custom_arguments import DEFAULT_LOG, LOG_ALWAYS
from . import util
from .util import shlex_join

__all__ = [
    "join_definitions",
    "split_statements_to_definitions",
    "PREFER_PASSING",
    "PASSING",
    "PREFER_NONPASSING",
    "NONPASSING",
    "get_preferred_passing",
    "get_fallback_passing",
]

PREFER_PASSING, PASSING, PREFER_NONPASSING, NONPASSING = (
    "prefer-passing",
    "passing",
    "prefer-nonpassing",
    "nonpassing",
)

DEFAULT_PREFERENCE_KEY = "parse_with"
_DEFAULT_PASSING = PREFER_NONPASSING


def get_preferred_fallback_passing(key, preference_key=DEFAULT_PREFERENCE_KEY, **env):
    preference = env.get(preference_key, _DEFAULT_PASSING)
    if preference in (PREFER_PASSING, PASSING, PREFER_NONPASSING, NONPASSING):
        default = (
            env.get("passing_" + key)
            if preference in (PREFER_PASSING, PASSING)
            else env.get(key)
        )
        fallback = (
            None
            if preference in (PASSING, NONPASSING)
            else env.get("passing_" + key)
            if preference == PREFER_NONPASSING
            else env.get(key)
        )
        return (default, fallback)
    raise ValueError(
        "Invalid value for %s: must be one of %s"
        % (
            preference_key,
            ", ".join(
                map(repr, (PREFER_PASSING, PASSING, PREFER_NONPASSING, NONPASSING))
            ),
        )
    )


def get_preferred_passing(key, **env):
    default, fallback = get_preferred_fallback_passing(key, **env)
    return default if default is not None else fallback


def get_fallback_passing(key, **env):
    default, fallback = get_preferred_fallback_passing(key, **env)
    # no fallback if we use fallback as default
    return fallback if default is not None else None


def get_definitions_diff(previous_definition_string, new_definition_string):
    """Returns a triple of lists (definitions_removed,
    definitions_shared, definitions_added)"""
    old_definitions = [i for i in previous_definition_string.split("|") if i]
    new_definitions = [i for i in new_definition_string.split("|") if i]
    if all(i == "branch" for i in old_definitions + new_definitions):  # work
        # around bug #5577 when all theorem names are "branch", we
        # don't assume that names are unique, and instead go by
        # ordering
        removed = []
        shared = []
        added = []
        for i in range(max((len(old_definitions), len(new_definitions)))):
            if i < len(old_definitions) and i < len(new_definitions):
                if old_definitions[i] == new_definitions[i]:
                    shared.append(old_definitions[i])
                else:
                    removed.append(old_definitions[i])
                    added.append(new_definitions[i])
            elif i < len(old_definitions):
                removed.append(old_definitions[i])
            elif i < len(new_definitions):
                added.append(new_definitions[i])
        return (tuple(removed), tuple(shared), tuple(added))
    else:
        return (
            tuple(i for i in old_definitions if i not in new_definitions),
            tuple(i for i in old_definitions if i in new_definitions),
            tuple(i for i in new_definitions if i not in old_definitions),
        )


def strip_newlines(string):
    if not string:
        return string
    if string[0] == "\n":
        return string[1:]
    if string[-1] == "\n":
        return string[:-1]
    return string


def split_statements_to_definitions(
    statements,
    log=DEFAULT_LOG,
    coqtop=("coqtop",),
    coqtop_args=tuple(),
    fallback_coqtop=None,
    fallback_coqtop_args=None,
    preference_key=DEFAULT_PREFERENCE_KEY,
    **kwargs,
):
    """Splits a list of statements into chunks which make up
    independent definitions/hints/etc."""
    if preference_key is not None:
        env = {"coqtop": coqtop, "coqtop_args": coqtop_args, "log": log}
        env.update(kwargs)
        coqtop, fallback_coqtop = get_preferred_fallback_passing(
            "coqtop", preference_key=preference_key, **env
        )
        coqtop_args, fallback_coqtop_args = get_preferred_fallback_passing(
            "coqtop_args", preference_key=preference_key, **env
        )
        return split_statements_to_definitions(
            statements,
            log=log,
            coqtop=coqtop,
            coqtop_args=coqtop_args,
            fallback_coqtop=fallback_coqtop,
            fallback_coqtop_args=fallback_coqtop_args,
            preference_key=None,
            **kwargs,
        )
    if coqtop is None and fallback_coqtop is not None:
        return split_statements_to_definitions(
            statements,
            log=log,
            coqtop=fallback_coqtop,
            coqtop_args=fallback_coqtop_args,
            fallback_coqtop=None,
            fallback_coqtop_args=None,
            preference_key=None,
            **kwargs,
        )

    def fallback():
        if fallback_coqtop is not None:
            log(
                "Your preferred version of coqtop (%s) doesn't support -emacs -time.  Falling back to fallback coqtop (%s)."
                % (shlex_join(coqtop), shlex_join(fallback_coqtop))
            )
            return split_statements_to_definitions(
                statements,
                log=log,
                coqtop=fallback_coqtop,
                coqtop_args=fallback_coqtop_args,
                fallback_coqtop=None,
                fallback_coqtop_args=None,
                preference_key=None,
                **kwargs,
            )
        else:
            log(
                "Your version of coqtop doesn't support -emacs -time.  Falling back to more error-prone method."
            )
            return split_definitions_old.split_statements_to_definitions(
                statements, log=log, coqtop=coqtop, coqtop_args=coqtop_args
            )

    # check for -time
    if not get_coq_accepts_time(coqtop, log=log) or not get_coq_accepts_emacs(
        coqtop, log=log
    ):
        return fallback()
    if not get_proof_term_works_with_time(coqtop, is_coqtop=True, log=log, **kwargs):
        statements = postprocess_split_proof_term(statements, log=log, **kwargs)
    assert isinstance(coqtop, tuple), coqtop
    p = Popen(
        [*coqtop, "-q", "-emacs", "-time"] + list(coqtop_args),
        stdout=PIPE,
        stderr=STDOUT,
        stdin=PIPE,
    )
    chars_time_reg = re.compile(
        r"\s*Chars ([0-9]+) - ([0-9]+) [^\s]+(?: [0-9\.]+ secs \([0-9\.]+u,[0-9\.]+s\)\s*)?".replace(
            " ", r"\s*"
        )
    )
    prompt_reg = re.compile(
        r"^(.*)<prompt>([^<]*?) < ([0-9]+) ([^<]*?) ([0-9]+) < ([^<]*)$".replace(
            " ", r"\s*"
        ),
        flags=re.DOTALL,
    )
    defined_reg = re.compile(
        r"^\s*(?:<infomsg>)?([^\s]+) is (?:defined|assumed|declared)(?:</infomsg>)?$",
        re.MULTILINE,
    )
    proof_using_reg = re.compile(
        r"^\s*<infomsg>\s*The proof of ([^\s]+) should start with(?: one of the following commands)?: ([^<]+)</infomsg>".replace(
            " ", r"\s+"
        ),
        flags=re.MULTILINE | re.DOTALL,
    )
    # goals and definitions are on stdout, prompts are on stderr
    statements_string = "Set Suggest Proof Using.\n" + "\n".join(statements) + "\n\n"
    statements_bytes = statements_string.encode("utf-8")
    log("Sending statements to coqtop...")
    log(statements_string, level=3)
    (stdout, stderr) = p.communicate(input=statements_bytes)
    stdout = util.s(stdout)
    if "know what to do with -time" in stdout.strip().split("\n")[0]:
        # we're using a version of coqtop that doesn't support -time
        return fallback()
    log("Done.  Splitting to definitions...")

    rtn = []
    cur_definition = {}
    cur_outputs = []
    last_definitions = "||"
    cur_definition_names = "||"
    last_char_end = 0

    # log('re.findall(' + repr(r'Chars ([0-9]+) - ([0-9]+) [^\s]+ (.*?)<prompt>([^<]*?) < ([0-9]+) ([^<]*?) ([0-9]+) < ([^<]*?)</prompt>'.replace(' ', r'\s*')) + ', ' + repr(stdout) + ', ' + 'flags=re.DOTALL)', level=3)
    for i, prompt in enumerate(stdout.split("</prompt>")):
        log("Processing:\n%s" % prompt, level=4)
        if prompt.strip() == "":
            continue
        times = chars_time_reg.findall(prompt)
        if len(times) == 0:
            if i == 0:
                log("Warning: ignoring %s" % repr(prompt), level=3)
            else:
                log("Warning: Could not find timing info in %s" % repr(prompt), level=1)
            continue
        elif len(times) > 1:
            log(
                "Warning: found multiple timing info in %s: %s"
                % (repr(prompt), repr(times)),
                level=1,
            )
        full_response_text = chars_time_reg.sub(" ", prompt)
        proof_using_match = proof_using_reg.findall(full_response_text)
        proof_using_options = []
        proof_using_thm_name = None
        if proof_using_match:
            full_response_text = proof_using_reg.sub(" ", full_response_text)
            if len(proof_using_match) > 1:
                log(
                    "Warning: found multiple proof using info in %s: %s"
                    % (prompt, str(proof_using_match)),
                    level=LOG_ALWAYS,
                )
            proof_using_thm_name, proof_using_cmds = proof_using_match[0]
            proof_using_options = [
                s.strip() for s in proof_using_cmds.split("Proof using") if s.strip()
            ]
            proof_using_options = [s.split(".")[0].strip() for s in proof_using_options]
            # log("Warning: found proof using info in %s: %s" % (repr(prompt), repr(proof_using_match)), level=1)
            # continue
        proof_using_options_used = False
        for char_start, char_end in times:
            char_start, char_end = int(char_start), int(char_end)
            # if we've travelled backwards in time, as in
            # COQBUG(https://github.com/coq/coq/issues/14475); we just
            # ignore this statement
            if char_end <= last_char_end:
                continue
            match = prompt_reg.match(full_response_text)
            if not match:
                log(
                    "Warning: Could not find statements in %d:%d: %s"
                    % (char_start, char_end, full_response_text),
                    level=LOG_ALWAYS,
                )
                continue
            (
                response_text,
                cur_name,
                line_num1,
                cur_definition_names,
                line_num2,
                unknown,
            ) = match.groups()
            statement = strip_newlines(
                statements_bytes[last_char_end:char_end].decode("utf-8")
            )
            last_char_end = char_end

            terms_defined = defined_reg.findall(response_text.strip())

            definitions_removed, definitions_shared, definitions_added = (
                get_definitions_diff(last_definitions, cur_definition_names)
            )

            # first, to be on the safe side, we add the new
            # definitions key to the dict, if it wasn't already there.
            if (
                cur_definition_names.strip("|")
                and cur_definition_names not in cur_definition
            ):
                cur_definition[cur_definition_names] = {
                    "statements": [],
                    "terms_defined": [],
                    "proof_using_options_map": [],
                    "proof_using_options": [],
                    "outputs": [],
                }

            log(
                (
                    statement,
                    (char_start, char_end),
                    definitions_removed,
                    proof_using_thm_name,
                    proof_using_options,
                    terms_defined,
                    "last_definitions:",
                    last_definitions,
                    "cur_definition_names:",
                    cur_definition_names,
                    cur_definition.get(last_definitions, []),
                    cur_definition.get(cur_definition_names, []),
                    response_text,
                ),
                level=2,
            )

            # first, we handle the case where we have just finished
            # defining something.  This should correspond to
            # len(definitions_removed) > 0 and len(terms_defined) > 0.
            # If only len(definitions_removed) > 0, then we have
            # aborted something (or we're in Coq >= 8.11; see
            # https://github.com/coq/coq/issues/15201).  If only
            # len(terms_defined) > 0, then we have defined something
            # with a one-liner.
            if definitions_removed:
                cur_definition[last_definitions]["statements"].append(statement)
                cur_definition[last_definitions]["terms_defined"] += terms_defined
                cur_definition[last_definitions]["outputs"].append(response_text)
                if proof_using_options:
                    cur_definition[last_definitions]["proof_using_options_map"].append(
                        (proof_using_thm_name, tuple(proof_using_options))
                    )
                    cur_definition[last_definitions]["proof_using_options"] += (
                        proof_using_options
                    )
                    proof_using_options = []
                if cur_definition_names.strip("|"):
                    # we are still inside a definition.  For now, we
                    # flatten all definitions.
                    #
                    # TODO(jgross): Come up with a better story for
                    # nested definitions.
                    cur_definition[cur_definition_names]["statements"] += (
                        cur_definition[last_definitions]["statements"]
                    )
                    cur_definition[cur_definition_names]["terms_defined"] += (
                        cur_definition[last_definitions]["terms_defined"]
                    )
                    cur_definition[cur_definition_names]["proof_using_options_map"] += (
                        cur_definition[last_definitions]["proof_using_options_map"]
                    )
                    cur_definition[cur_definition_names]["proof_using_options"] += (
                        cur_definition[last_definitions]["proof_using_options"]
                    )
                    cur_definition[cur_definition_names]["outputs"] += cur_definition[
                        last_definitions
                    ]["outputs"]
                    del cur_definition[last_definitions]
                else:
                    # we're at top-level, so add this as a new
                    # definition
                    rtn.append(
                        {
                            "statements": tuple(
                                cur_definition[last_definitions]["statements"]
                            ),
                            "statement": "\n".join(
                                cur_definition[last_definitions]["statements"]
                            ),
                            "terms_defined": tuple(
                                cur_definition[last_definitions]["terms_defined"]
                            ),
                            "proof_using_options_map": tuple(
                                cur_definition[last_definitions][
                                    "proof_using_options_map"
                                ]
                            ),
                            "proof_using_options": tuple(
                                cur_definition[last_definitions]["proof_using_options"]
                            ),
                            "outputs": tuple(
                                cur_definition[last_definitions]["outputs"]
                            ),
                            "output": "\n".join(
                                cur_definition[last_definitions]["outputs"]
                            ),
                        }
                    )
                    del cur_definition[last_definitions]
                    # print('Adding:')
                    # print(rtn[-1])
            elif terms_defined:
                if cur_definition_names.strip("|"):
                    # we are still inside a definition.  For now, we
                    # flatten all definitions.
                    #
                    # TODO(jgross): Come up with a better story for
                    # nested definitions.
                    cur_definition[cur_definition_names]["statements"].append(statement)
                    cur_definition[cur_definition_names]["terms_defined"] += (
                        terms_defined
                    )
                    cur_definition[cur_definition_names]["outputs"].append(
                        response_text
                    )
                else:
                    # we're at top level, so add this as a new
                    # definition
                    rtn.append(
                        {
                            "statements": (statement,),
                            "statement": statement,
                            "terms_defined": tuple(terms_defined),
                            "proof_using_options_map": (
                                ((proof_using_thm_name, tuple(proof_using_options)),)
                                if proof_using_options
                                else ()
                            ),
                            "proof_using_options": tuple(proof_using_options),
                            "outputs": (response_text,),
                            "output": response_text,
                        }
                    )
                    proof_using_options_used = True

            # now we handle the case where we have just opened a fresh
            # definition.  We've already added the key to the
            # dictionary.
            elif definitions_added:
                # print(definitions_added)
                cur_definition[cur_definition_names]["statements"].append(statement)
                cur_definition[cur_definition_names]["outputs"].append(response_text)
            else:
                # if we're in a definition, append the statement to
                # the queue, otherwise, just add it as it's own
                # statement
                if cur_definition_names.strip("|"):
                    cur_definition[cur_definition_names]["statements"].append(statement)
                    cur_definition[cur_definition_names]["outputs"].append(
                        response_text
                    )
                else:
                    rtn.append(
                        {
                            "statements": (statement,),
                            "statement": statement,
                            "terms_defined": (),
                            "proof_using_options_map": (
                                ((proof_using_thm_name, tuple(proof_using_options)),)
                                if proof_using_options
                                else ()
                            ),
                            "proof_using_options": tuple(proof_using_options),
                            "outputs": (response_text,),
                            "output": response_text,
                        }
                    )
                    proof_using_options_used = True

            last_definitions = cur_definition_names
        if proof_using_options and not proof_using_options_used:
            log(
                "Warning: Unused proof using options: %s in %s"
                % (str(proof_using_options), prompt),
                level=LOG_ALWAYS,
            )
    log((last_definitions, cur_definition_names), level=2)
    if last_definitions.strip("||"):
        rtn.append(
            {
                "statements": tuple(cur_definition[cur_definition_names]["statements"]),
                "statement": "\n".join(
                    cur_definition[cur_definition_names]["statements"]
                ),
                "terms_defined": tuple(
                    cur_definition[cur_definition_names]["terms_defined"]
                ),
                "proof_using_options_map": tuple(
                    cur_definition[cur_definition_names]["proof_using_options_map"]
                ),
                "proof_using_options": tuple(
                    cur_definition[cur_definition_names]["proof_using_options"]
                ),
                "outputs": tuple(cur_definition[cur_definition_names]["outputs"]),
                "output": "\n".join(cur_definition[cur_definition_names]["outputs"]),
            }
        )
        del cur_definition[last_definitions]

    if last_char_end + 1 < len(statements_bytes):
        last_statement = statements_bytes[last_char_end:].decode("utf-8")
        log(
            "Appending end of code from %d to %d: %s"
            % (last_char_end, len(statements_bytes), last_statement),
            level=2,
        )
        last_statement = strip_newlines(last_statement)
        rtn.append(
            {
                "statements": tuple(
                    last_statement,
                ),
                "statement": last_statement,
                "terms_defined": (),
                "proof_using_options_map": (
                    ((proof_using_thm_name, tuple(proof_using_options)),)
                    if proof_using_options and not proof_using_options_used
                    else ()
                ),
                "proof_using_options": (
                    tuple(proof_using_options)
                    if proof_using_options and not proof_using_options_used
                    else ()
                ),
                "outputs": (),
                "output": "",
            }
        )
    if rtn[0]["statement"].strip() == "Set Suggest Proof Using.":
        rtn = rtn[1:]

    return rtn


def join_definitions(definitions):
    return "\n".join(i["statement"] for i in definitions)


def split_statements_to_definitions_with_options(
    statements,
    **kwargs,
):
    new_statements = []
    for statement in statements:
        # work around COQBUG(https://github.com/rocq-prover/rocq/issues/21091)
        parts = statement.strip().split()
        if not (len(parts) >= 2 and parts[0] == "Proof" and parts[1].startswith("(")):
            new_statements.append("Print Options.")
        new_statements.append(statement)
    new_statements.append("Print Options.")
    definitions = split_statements_to_definitions(new_statements, **kwargs)
    new_definitions = []
    cur_options = set()
    for definition in definitions:
        if definition["statement"].strip() == "Print Options.":
            new_options = parse_options_settings(definition["output"])
            if new_definitions and "new_options" not in new_definitions[-1]:
                new_definitions[-1]["new_options"] = tuple(
                    opt for opt in new_options if opt not in cur_options
                )
            cur_options = set(new_options)
        else:
            if len(definition["statements"]) == len(definition["outputs"]):
                statement_outputs = [(stmt, output) for stmt, output in zip(definition["statements"], definition["outputs"]) if stmt.strip() != "Print Options."]
                definition["statements"] = tuple(stmt for stmt, _ in statement_outputs)
                definition["outputs"] = tuple(output for _, output in statement_outputs)
            else:
                definition["statements"] = tuple(
                    stmt
                    for stmt in definition["statements"]
                    if stmt.strip() != "Print Options."
                )
            definition["statement"] = "\n".join(definition["statements"])
            definition["output"] = "\n".join(definition["outputs"])
            new_definitions.append(definition)
    return new_definitions



import os
import tempfile
import time
from hashlib import sha256

from . import util
from .memoize import memoize

__all__ = [
    "clean_v_file",
    "clean_extra_coq_files",
    "write_bytes_to_file",
    "read_bytes_from_file",
    "write_to_file",
    "read_from_file",
    "restore_file",
    "write_bytes_to_file_or_shorten_name",
    "write_to_file_or_shorten_name",
]


def clean_extra_coq_files(v_file_name, extra_exts=tuple()):
    for pre in ("", "."):
        for ext in tuple(
            list(extra_exts) + [".glob", ".vo", ".d", ".v.d", ".aux", ".vos", ".vok"]
        ):
            name = "".join(
                (
                    os.path.dirname(v_file_name[:-2]),
                    os.sep,
                    pre,
                    os.path.basename(v_file_name[:-2]),
                    ext,
                )
            )
            if os.path.exists(name):
                os.remove(name)


def clean_v_file(file_name):
    clean_extra_coq_files(file_name, extra_exts=(".v",))


def backup(file_name, ext=".bak"):
    if not ext:
        raise ValueError
    if os.path.exists(file_name):
        backup(file_name + ext, ext=ext)
        os.rename(file_name, file_name + ext)


@memoize
def memobackup(file_name, ext=".bak"):
    return backup(file_name, ext)


FILE_CACHE = {}


def write_bytes_to_file(
    file_name, contents, do_backup=False, backup_ext=".bak", memoize=False
):
    assert contents is bytes(contents)
    backed_up = False
    while not backed_up:
        try:
            if do_backup:
                if memoize:
                    memobackup(file_name, ext=backup_ext)
                else:
                    backup(file_name, ext=backup_ext)
            backed_up = True
        except IOError as e:
            print(
                "Warning: f.write(%s) failed with %s\nTrying again in 10s"
                % (file_name, repr(e))
            )
            time.sleep(10)
    written = False
    while not written:
        try:
            with tempfile.NamedTemporaryFile(
                mode="wb",
                dir=os.path.dirname(file_name),
                prefix=os.path.basename(file_name) + ".",
                delete=False,
            ) as f:
                f.write(contents)
                temp_file_name = f.name
            os.rename(temp_file_name, file_name)
            written = True
            FILE_CACHE[file_name] = (os.stat(file_name).st_mtime, contents)
        except IOError as e:
            if e.errno == 36:  # file name too long
                raise e
            print(
                "Warning: f.write(%s) failed with %s\nTrying again in 10s"
                % (file_name, repr(e))
            )
            time.sleep(10)


def write_to_file(file_name, contents, *args, **kwargs):
    return write_bytes_to_file(
        file_name, contents.replace("\n", os.linesep).encode("utf-8"), *args, **kwargs
    )


def make_shorter_file_name(file_name):
    file_dir, file_name = os.path.split(file_name)
    file_name, output_file_ext = os.path.splitext(file_name)
    file_hash = sha256(file_name.encode("utf-8")).hexdigest()
    file_hash = f"...{file_hash}..."
    if len(file_name + output_file_ext) > 255:
        max_len = 255
    else:
        max_len = int(0.75 * len(file_name))
    max_len = max_len - len(output_file_ext) - len(file_hash)
    file_name = f"{file_name[: max_len // 2]}{file_hash}{file_name[-max_len // 2 :]}{output_file_ext}"
    return os.path.join(file_dir, file_name)


def write_bytes_to_file_or_shorten_name(file_name, contents, *args, **kwargs):
    while True:
        try:
            write_bytes_to_file(file_name, contents, *args, **kwargs)
            return file_name
        except OSError as e:
            if e.errno != 36:
                raise e
            print(
                f"Warning: {file_name} is too long, so we're trying again with a shorter name"
            )
            file_name = make_shorter_file_name(file_name)


def write_to_file_or_shorten_name(file_name, contents, *args, **kwargs):
    return write_bytes_to_file_or_shorten_name(
        file_name, contents.replace("\n", os.linesep).encode("utf-8"), *args, **kwargs
    )


def restore_file(file_name, backup_ext=".bak", backup_backup_ext=".unbak"):
    if not os.path.exists(file_name + backup_ext):
        raise IOError
    if os.path.exists(file_name):
        if backup_backup_ext:
            backup(file_name, backup_backup_ext)
        else:
            os.remove(file_name)
    os.rename(file_name + backup_ext, file_name)


def read_bytes_from_file(file_name):
    if (
        file_name in FILE_CACHE.keys()
        and os.stat(file_name).st_mtime == FILE_CACHE[file_name][0]
    ):
        return FILE_CACHE[file_name][1]
    with open(file_name, "rb") as f:
        FILE_CACHE[file_name] = (os.stat(file_name).st_mtime, f.read())
        return FILE_CACHE[file_name][1]


def read_from_file(file_name):
    return util.normalize_newlines(read_bytes_from_file(file_name).decode("utf-8"))



# from http://stackoverflow.com/questions/375427/non-blocking-read-on-a-subprocess-pipe-in-python

import sys
from subprocess import PIPE, STDOUT
import subprocess, time
from threading import Thread

__all__ = ["Popen_async", "Empty", "PIPE", "STDOUT"]

try:
    from Queue import Queue, Empty
except ImportError:
    from queue import Queue, Empty  # python 3.x

ON_POSIX = "posix" in sys.builtin_module_names


def enqueue_output(out, queue):
    for line in iter((lambda: out.read(1)), b""):
        # print('0: %s' % line)
        queue.put(line)
    out.close()


class Popen_async(object):
    """Allows non-blocking reading from a Popen, by

    p = Popen_async(['myprogram.exe'], stdout=PIPE, stderr=PIPE)
    try:  line = p.stdout.get_nowait() # or p.stdout.get(timeout=.1)
    except Empty:
        print('no output yet')
    else: # got line
    """

    def __init__(self, *args, **kwargs):
        self._p = subprocess.Popen(*args, bufsize=1, close_fds=ON_POSIX, **kwargs)
        self.stdout = Queue()
        self.stderr = Queue()
        self._tout = Thread(target=enqueue_output, args=(self._p.stdout, self.stdout))
        # self._terr = Thread(target=enqueue_output, args=(self._p.stderr, self.stderr))
        self._tout.daemon = True  # thread dies with the program
        # self._terr.daemon = True # thread dies with the program
        self._tout.start()
        # self._terr.start()
        self.stdin = self._p.stdin
        time.sleep(0.1)



import subprocess, re
from .strip_comments import strip_comments
from .custom_arguments import DEFAULT_LOG
from .coq_version import get_coq_accepts_time
from . import util
from .util import PY3

if PY3:
    from .util import raw_input

__all__ = [
    "split_coq_file_contents",
    "split_coq_file_contents_with_comments",
    "get_coq_statement_byte_ranges",
    "UnsupportedCoqVersionError",
    "postprocess_split_proof_term",
    "split_leading_comments_and_whitespace",
]


def fill_kwargs(kwargs):
    ret = {
        "log": DEFAULT_LOG,
        "coqc": "coqc",
        "coqc_args": tuple(),
    }
    ret.update(kwargs)
    return ret


class UnsupportedCoqVersionError(Exception):
    pass


def merge_quotations(statements, sp=" "):
    """If there are an odd number of "s in a statement, assume that we
    broke the middle of a string.  We recombine that string."""

    cur = None
    for i in statements:
        if i.count('"') % 2 != 0:
            if cur is None:
                cur = i
            else:
                yield (cur + sp + i)
                cur = None
        elif cur is None:
            yield i
        else:
            cur += sp + i


BRACES = "{}-*+"


def split_leading_braces(statement):
    while statement.strip() and statement.strip()[0] in BRACES:
        idx = statement.index(statement.strip()[0]) + 1
        first, rest = statement[:idx], statement[idx:]
        # if 'not_reachable_iff' in statement or 'not_reachable_iff' in first:
        #    print((first, rest))
        yield first
        statement = rest
    if statement:
        yield statement


def split_leading_comments_and_whitespace(text):
    comment_level = 0
    in_str = False
    skip = False
    for i, (ch, next_ch) in enumerate(zip(text, text[1:] + "\0")):
        if skip:
            skip = False
        elif in_str:
            if ch in '"':
                in_str = False
        elif ch == "(" and next_ch == "*":
            comment_level += 1
            skip = True
        elif ch == "*" and next_ch == ")":
            comment_level -= 1
            skip = True
        elif comment_level > 0:
            if ch in '"':
                in_str = True
        elif ch.isspace():
            pass
        else:
            return text[:i], text[i:]
    return text, ""


def split_merge_comments(statements):
    """Find open and close comments."""

    def is_genuine_ending(cur):
        return (
            (cur.strip()[:2] == "(*" and cur.strip()[-2:] == "*)")  # cur is just a comment
            or cur.strip()[-1:] in ".}"
            or cur.strip() in BRACES
        )

    cur = None
    comment_level = 0
    for stmt in statements:
        str_count = 0
        for i in re.split(r"(?<=\*\)) ", stmt.replace("*)", "*) ")):
            if str_count % 2 == 0:
                i2 = re.sub('"[^"]*"|"[^"]*$', "", i)
            else:
                i2 = re.sub('^[^"]*"|"[^"]*"', "", i)
            str_count += i.count('"')
            # print((repr(i), comment_level, repr(i2)))
            if "(*" not in i2 and "*)" not in i2:
                if comment_level < 0:
                    if cur is not None:
                        for ret in split_leading_braces(cur + i):
                            yield ret
                        cur = None
                    else:
                        raw_input("UNEXPECTED COMMENT: %s" % i)
                        yield i
                    comment_level = 0
                elif comment_level == 0:
                    if cur is None:
                        for ret in split_leading_braces(i):
                            yield ret
                    else:
                        cur += i
                elif cur is None:
                    cur = i
                else:
                    cur += i
            else:
                comment_level += i2.count("(*") - i2.count("*)")
                if cur is None:
                    cur = i
                else:
                    cur += i
            if cur is not None:
                curs = list(split_leading_braces(cur))
                while curs and curs[0].strip() in BRACES:
                    yield curs.pop(0)
                cur = "".join(curs) if curs else None
            if (
                cur is not None and is_genuine_ending(cur) and comment_level == 0
            ):  # clear out cur, if we're done with it
                yield cur
                cur = None
    if cur is not None:
        print("Unterminated comment")
        yield cur


def split_coq_file_contents(contents):
    """Splits the contents of a coq file into multiple statements.

    This is done by finding one or three periods followed by
    whitespace.  This is a dumb algorithm, but it seems to be (nearly)
    the one that ProofGeneral and CoqIDE use.

    We additionally merge lines inside of quotations."""
    return list(merge_quotations(re.split(r"(?<=[^\.]\.\.\.)\s|(?<=[^\.]\.)\s", strip_comments(contents))))


def split_coq_file_contents_with_comments(contents):
    statements = re.split(
        r"(?<=[^\.]\.\.\.) |(?<=[^\.]\.) ", re.sub(r"((?<=[^\.]\.\.\.)\s|(?<=[^\.]\.)\s)", r" \1", contents)
    )
    # if contents != ''.join(statements):
    #    print('Splitting failed (initial split)!')
    # qstatements = list(merge_quotations(statements, sp=''))
    # if contents != ''.join(qstatements):
    #    print('Splitting failed (quote merge)!')
    # cstatements = list(split_merge_comments(qstatements))
    # if contents != ''.join(cstatements):
    #    print('Splitting failed (comment merge)!')
    return list(split_merge_comments(merge_quotations(statements, sp="")))


PROOF_TERM_REG = re.compile(r"(\s*)Proof(\s+)(.*?)\.(\s*)", flags=re.DOTALL | re.MULTILINE)


def postprocess_split_proof_term_iter(statements, on_first_example=None):
    """Returns an iterator for the statements passed which changes [Proof
    term.] into [Proof. exact (term). Qed.] unless "term" begins with
    "with" or "using"."""
    for statement_num, statement in enumerate(statements):
        match = PROOF_TERM_REG.match(statement)
        if match:
            indent, after_proof_space, proof_term, post_space = match.groups()
            if re.split(r"\s", proof_term)[0] not in (
                "using",
                "with",
            ):  # Proof using ... and Proof with ... are special
                if on_first_example:
                    on_first_example(statement_num, statement, proof_term)
                    on_first_example = None
                yield indent + "Proof."
                yield after_proof_space + "exact (" + proof_term + ")."
                yield " Qed." + post_space
            else:
                yield statement
        else:
            yield statement


def postprocess_split_proof_term(statements, do_warn=True, **kwargs):
    def do_warn_method(statement_num, statement, proof_term):
        kwargs["log"](
            """Warning: Your version of Coq suffers from bug #5349 (https://coq.inria.fr/bugs/show_bug.cgi?id=5349)
and does not support [Proof (term).] with -time.  Falling back to
replacing [Proof (term).] with [Proof. exact (term). Qed.], which may fail."""
        )

    if do_warn and "log" in kwargs.keys():
        do_warn = do_warn_method
    else:
        do_warn = None
    return list(postprocess_split_proof_term_iter(statements, do_warn))


RANGE_REG = re.compile(r"Chars ([0-9]+) - ([0-9]+) [^\s]+", flags=re.DOTALL)


def get_coq_statement_byte_ranges(file_name, coqc, **kwargs):
    kwargs = fill_kwargs(kwargs)
    if not get_coq_accepts_time(coqc, **kwargs):
        raise UnsupportedCoqVersionError

    assert isinstance(coqc, tuple), coqc
    p = subprocess.Popen(
        [*coqc, "-q", "-time"] + list(kwargs["coqc_args"]) + [file_name],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        stdin=subprocess.PIPE,
    )
    (stdout, stderr) = p.communicate()

    ranges = tuple((int(start), int(finish)) for start, finish in RANGE_REG.findall(util.s(stdout)))
    return ranges



from __future__ import division
from .memoize import memoize

__all__ = ["run_binary_search"]


@memoize
def apply_as_many_times_as_possible(f, x):
    if x is None:
        return None
    fx = apply_as_many_times_as_possible(f, f(x))
    if fx is None:
        return x
    return fx


def make_states(init, step):
    init = step(init)
    while init != None:
        yield init
        init = step(init)


def binary_search(f, ls):
    start, end = 0, len(ls)
    # assumption: forall i, not f(ls[i]) -> all(map(not f, ls[i:]))
    # invariant: all(map(not f, ls[end:])) and all(map(f, ls[:start]))
    mid = (start + end) // 2
    while mid < end:
        if f(ls[mid]):
            start = mid + 1
        else:
            end = mid
        mid = (start + end) // 2
    return start - 1


def run_binary_search(initial_state, check_state, step_state, save_good_state, valid_nondefault_actions):
    """
    Runs a binary search on initial_state to find the best final state.

    - initial_state : State
        Starting point.  Must be pure (no mutation necessary)

    - check_state : State -> bool
        returns True if the state is a good state, and False if it is bad.
        If run_binary_search returns any non-none value and
        check_state is pure, then check_state is guaranteed to return
        a truthy value on the returned state.

    - step_state : State -> Maybe Action -> State
        Takes a step based on the action, or returns None if no such step
        is possible.  The action is either None, or one from
        valid_nondefault_actions.  If no step is possible, step_state
        must return None, rather than its input state (otherwise
        run_binary_search may loop).  This argument must be a pure function.

        Requirement: check_state(st) == check_state(step_state(st, None))
        assuming step_state(st, None) != None

    - save_good_state : State -> ()
        Called after every discovery of a state that check_state is happy with

    - valid_nondefault_actions : [State]
        A list or tuple of actions.  Earlier actions are assumed to be better.

    Returns the state that is "best", as determined by a lexicographic
    ordering on the actions taken, where earlier actions in
    valid_nondefault_actions are better than later ones.
    """
    valid_actions = list(valid_nondefault_actions) + [None]

    @memoize
    def is_good(state):
        return check_state(state)

    def step_none(st):
        return step_state(st, None)

    def step_default(st):
        for action in valid_actions:
            try:
                return step_state(st, action)
            except ValueError:
                pass
        return step_none(st)

    last_good = None
    cur = initial_state
    while is_good(cur) and last_good != cur:
        last_good = cur
        current_states = list(make_states(cur, step_default))
        if current_states and is_good(current_states[0]):
            idx = binary_search(is_good, current_states)
            cur = current_states[idx]
            save_good_state(cur)
        else:
            for action in valid_actions:
                try:
                    next_st = step_state(cur, action)
                except ValueError:
                    continue
                if next_st and is_good(next_st):
                    cur = next_st
                    save_good_state(cur)
                    break
    return last_good
